/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: calculate.c,v 1.1.1.1 1999/11/05 01:55:18 shinmaru Exp $
 */

#include "mahjong.h"

static int calculate(int immaturities, int singletons, int no_uses, int pairs)
{
  int going = immaturities;
  
  if (immaturities <= no_uses) {
    no_uses -= immaturities;
  } else {
    immaturities -= no_uses;
    no_uses = 0;
    
    if (immaturities <= singletons) {
      singletons -= immaturities;
    } else {
      immaturities -= singletons;
      singletons = 0;
      going -= immaturities / 3;
    }
  }
  if (pairs == 0) {
    if (singletons > 0) {
      going += 1;
      if (no_uses > 0) {
        no_uses -= 1;
        singletons = max(0, singletons - 1);
      }
      else
        singletons = max(0, singletons - 2);
    }
  }
  singletons = (singletons + 2) / 3 * 2;
  return going + max(no_uses, singletons);
}

int calculate_going(set_t *sets, int count)
{
  int singletons = 0, pairs = 0, immaturities = 0;
  int going;
  int i;
  for (i = 0; i < count; i++) {
    set_t set = sets[i];
    if (!is_revealed(set)) {
      switch (type_of(set)) {
      case SINGLETON: singletons++; break;
      case PAIR:
        if (pairs++ == 0)
          break;
        /* ȴ */
      case CONJUNCT:
      case SPLITTER: immaturities++; break;
      }
    }
  }
  going = calculate(immaturities, singletons, 0, pairs);
  
  return going;
}

/* 䥪 */
int estimate_no_terminals(evaluated_hand_t *e)
{
  int value = 0;
  int i;
  for (i = 1; i < 8; i++) {
    value += e->singletons[i] + e->singletons[i + 9] + e->singletons[i + 18];
    value += e->triplets[i] + e->triplets[i + 9] + e->triplets[i + 18];
  }
  for (i = 1; i < 6; i++) {
    value += e->sequences[i] + e->sequences[i + 9] + e->sequences[i + 18];
  }
  value += min(2, e->sequences[ 6]);
  value += min(2, e->sequences[15]);
  value += min(2, e->sequences[24]);
  return value;
}

/*  */
int estimate_terminals(evaluated_hand_t *e)
{
  int value = 0;
  int i;
  for (i = 0; i < 27; i += 9) {
    value += e->singletons[i] + e->singletons[i + 8];
    value += e->triplets[i] + e->triplets[i + 8];
    value += e->sequences[i] + e->sequences[i + 6];
  }
  for (i = 27; i < MAX_SORT; i++) {
    value += e->singletons[i] + e->triplets[i];
  }
  return value;
}

/* ʿ */
int estimate_sequences(evaluated_hand_t *e)
{
  int singletons = 0, pairs = 0, conjuncts = 0, terminals = 0;
  int immaturities = 0, no_uses = 0;
  int seat_wind = seat_wind_of(e->pid);
  int going;
  int i;
  for (i = 0; i < e->count; i++) {
    set_t set = e->sets[i];
    int n = tile_of(set);
    switch (type_of(set)) {
    case PAIR:
      if (is_dragon(n) || n == seat_wind || n == round_wind())
        no_uses += 2; /* ƬˤϤʤʤ */
      else
        pairs++;
      break;
    case CONJUNCT:
      if (number_of(n) != 0 && number_of(n) != 7)
        conjuncts++;
      else
        terminals++;
      break;
    case SPLITTER: immaturities++; break;
    case SINGLETON:
    case TRIPLET:
      if (is_dragon(n) || n == seat_wind || n == round_wind())
        no_uses += tile_count(set); /* ʿ¤ˤϻȤʤ */
      else
        singletons += tile_count(set);
      break;
    case QUADRUPLET:
      return 0;
    }
  }
  if (pairs > 1) {
    singletons += (pairs - 1) * 2;
    pairs = 1;
  }
  immaturities += conjuncts + terminals;
  going = calculate(immaturities, singletons, no_uses, pairs);
  if (going == 1) {
    /* İפξԤǤʤФʤʤ */
    if (terminals != 0) return 0;
    if (conjuncts == 0) return 13 - going;
  }
  return 14 - going;
}

/* ָ */
int estimate_sisters(evaluated_hand_t *e)
{
  int value = 0;
  int i;
  for (i = 0; i < 27; i++) {
    if (e->sequences[i] > 3)
      value += e->sequences[i];
  }
  return value;
}

/* 쵤̴ */
int estimate_straight(evaluated_hand_t *e)
{
  int p0 = min(3, e->sequences[ 0]) + min(3, e->sequences[ 3]) + min(3, e->sequences[ 6]);
  int p1 = min(3, e->sequences[ 9]) + min(3, e->sequences[12]) + min(3, e->sequences[15]);
  int p2 = min(3, e->sequences[18]) + min(3, e->sequences[21]) + min(3, e->sequences[24]);
  return max(p0, max(p1, p2));
}

/* Ʊ */
int estimate_3_suit_equals(evaluated_hand_t *e)
{
  int value = 0;
  int i;
  for (i = 0; i < 7; i++) {
    int n1 = i, n2 = i + 9, n3 = i + 18;
    int q1 = min(3, e->sequences[n1]);
    int q2 = min(3, e->sequences[n2]);
    int q3 = min(3, e->sequences[n3]);
    int p = (q1 ? 1 : 0) | (q2 ? 2 : 0) | (q3 ? 4 : 0);
    int temp = 0;
    int s1, s2, s3;
    if (p != 7) {
      s1 = e->singletons[n1] + e->singletons[n1 + 1] + e->singletons[n1 + 2];
      s2 = e->singletons[n2] + e->singletons[n2 + 1] + e->singletons[n2 + 2];
      s3 = e->singletons[n3] + e->singletons[n3 + 1] + e->singletons[n3 + 2];
    }
    switch (p) {
    case 0: temp = s1 + s2 + s3; break;
    case 1: temp = q1 + s2 + s3; break;
    case 2: temp = s1 + q2 + s3; break;
    case 3: temp = q1 + q2 + s3; break;
    case 4: temp = s1 + s2 + q3; break;
    case 5: temp = q1 + s2 + q3; break;
    case 6: temp = s1 + q2 + q3; break;
    case 7: temp = q1 + q2 + q3; break;
    }
    if (value < temp)
      value = temp;
  }
  return value;
}

/* 쿧 */
int estimate_1_suit(evaluated_hand_t *e)
{
  int value = 0;
  int i, end;

  if (e->one_suit >= 0) {
    for (i = 9 * e->one_suit, end = 9 + i; i < end; i++) {
      value += e->singletons[i] + e->triplets[i] + e->sequences[i];
    }
    for (i = 27; i < MAX_SORT; i++) {
      value += e->singletons[i] + e->triplets[i];
    }
  }
  return value;
}

/* С */
int estimate_triplets(evaluated_hand_t *e)
{
  int value = 0;
  int i;
  for (i = 0; i < MAX_SORT; i++) {
    value += e->triplets[i];
  }
  return value;
}

/* л */
int estimate_7_pairs(evaluated_hand_t *e)
{
  int pairs = 0;
  int i;
  for (i = 0; i < e->count; i++) {
    if (type_of(e->sets[i]) == PAIR)
      pairs++;
  }
  return pairs * 2;
}

/* ̵ */
int estimate_13_orphans(evaluated_hand_t *e)
{
  int value = 0;
  int i;
  for (i = 0; i < 13; i++) {
    value += e->singletons[terminal_tiles[i]];
  }
  return value;
}
