/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: debug.c,v 1.2 1999/11/06 01:07:49 shinmaru Exp $
 */

#include "mahjong.h"

void print_hand(int pid)
{
  player_t *p = &players[pid];
  int i, j;
  
  for (i = 0; i < MAX_SORT; i++) {
    for (j = p->hand[i]; j > 0; j--) {
      printf("%s", tile_face[i]);
    }
  }
  for (i = 0; i < p->revealed_count; i++) {
    print_set(p->revealed[i]);
  }
}

void print_evaluated_hand(evaluated_hand_t *e)
{
  int i;
  for (i = 0; i < e->count; i++) {
    print_set(e->sets[i]);
  }
  printf(": going = %d\n", e->going);
  for (i =  0; i <  9; i++) { printf("%d ", e->singletons[i]); } putchar(':');
  for (i =  0; i <  9; i++) { printf("%d ", e->sequences[i]);  } putchar(':');
  for (i =  0; i <  9; i++) { printf("%d ", e->triplets[i]);   } putchar('\n');
  for (i =  9; i < 18; i++) { printf("%d ", e->singletons[i]); } putchar(':');
  for (i =  9; i < 18; i++) { printf("%d ", e->sequences[i]);  } putchar(':');
  for (i =  9; i < 18; i++) { printf("%d ", e->triplets[i]);   } putchar('\n');
  for (i = 18; i < 27; i++) { printf("%d ", e->singletons[i]); } putchar(':');
  for (i = 18; i < 27; i++) { printf("%d ", e->sequences[i]);  } putchar(':');
  for (i = 18; i < 27; i++) { printf("%d ", e->triplets[i]);   } printf("\n\n");
}

/* ξ̤ */
static void print_conjuncts(hook_table_t h, int suit)
{
  int start = suit * 9;
  int end = 6 + suit * 9;
  int test_bits = bit(HOOK_BELOW) | bit(HOOK_ABOVE);
  int n;
  for (n = start; n < end; n++) {
    if ((h[n].types & test_bits) == bit(HOOK_BELOW)) {
      printf(number_of(n) < 3 && bittest(h[n + 6].types, HOOK_ABOVE)
             ? " %s%s%s" : " %s%s",
             tile_face[n], tile_face[n + 3], tile_face[n + 6]);
    }
  }
}

/* ĥ */
static void print_terminals(hook_table_t h, int suit)
{
  int n1 = 2 + 9 * suit;
  int n2 = 6 + 9 * suit;
  int cnt = 0;
  if (bittest(h[n1].types, HOOK_ABOVE))
    printf(" %s", tile_face[n1]), cnt++;
  if (bittest(h[n2].types, HOOK_BELOW))
    printf(cnt ? "%s" : " %s", tile_face[n2]);
}

/* ĥ */
static void print_splitters(hook_table_t h, int suit)
{
  int start = 1 + suit * 9;
  int end = 8 + suit * 9;
  int cnt = 0;
  int n;
  for (n = start; n < end; n++)
    if (bittest(h[n].types, HOOK_MIDDLE)) {
      if (cnt++ == 0)
        printf(" ");
      printf("%s", tile_face[n]);
    }
}

static void print_any(hook_table_t h, int type, const char *caption)
{
  int cnt = 0;
  int n;
  for (n = 0; n < MAX_SORT; n++) {
    if (bittest(h[n].types, type)) {
      if (cnt++ == 0)
        printf(caption);
      printf("%s", tile_face[n]);
    }
  }
}

/* Ԥ׾ */
void print_hook_table(hook_table_t hook)
{
  int suit;
  for (suit = 0; suit < 3; suit++) {
    print_conjuncts(hook, suit);
    print_terminals(hook, suit);
    print_splitters(hook, suit);
  }
  print_any(hook, HOOK_PAIR, " ñ");
  print_any(hook, HOOK_TRIPLET, " ");
  print_any(hook, HOOK_SINGLETON, " ");
}

/* ׾ */
void print_going_table(void)
{
  int i;
  printf("\n                ʿ       \n");
  for (i = 0; i < MAX_SORT; i++) {
    int gone = going_table[i].max_gone;
    if (gone > 0) {
      int j;
      int going = 14 - gone;
      printf(going > 1 ? "%s[%s](%dİ):" : "%s[%s](İ) :",
             going_table[i].quadrupling ? "" : "",
             tile_face[i], going - 1);
      for (j = 0; j < MAX_EVAL; j++) {
        printf("%2d ", going_table[i].values[j]);
      }
      print_hook_table(going_table[i].hook);
      
#ifndef PRINT_SINGLETONS
      putchar('\n');
#else
      for (j = 0, cnt = 0; j < MAX_SORT; j++) {
        if (going_table[i].singletons[j]) {
          if (cnt++ == 0)
            printf(" [");
          printf("%s", tile_face[j]);
        }
      }
      printf(cnt ? "]\n" : "\n");
#endif
    }
  }
}

/* פɾͤɽ */
void print_discards(int *dead_tiles)
{
  int i, n;
  for (i = 0; i < MAX_SORT; i++) {
    int total = 0;
    if (going_table[i].max_gone > 0) {
      for (n = 0; n < MAX_SORT; n++)
        if (going_table[i].hook[n].types != 0) {
          int rest_tiles = 4 - dead_tiles[n];
          if (rest_tiles >= going_table[i].hook[n].needs)
            total += rest_tiles;
        }
      printf("%s:%d ", tile_face[i], total);
    }
  }
  putchar('\n');
}

/* ɽ */
void print_scoresheet(scoresheet_t *sheet)
{
  hand_name_t *hand;

  if (sheet == NULL) {
    return;
  }
  for (hand = sheet->names; hand != NULL; hand = hand->next) {
    printf(hand->bonus > 0 ? "%-32s%d\n" : "%-32s\n", hand->name, hand->bonus);
  }
  printf(sheet->fmt, sheet->point, sheet->bonus);
  printf(" %d\n", sheet->total);
}
