/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: deck.c,v 1.2 1999/11/06 01:07:49 shinmaru Exp $
 */

#include "mahjong.h"

#define WRAP_COUNT		8	/* Τפޤ֤ */

#define MAX_BONE		512	/* Ѳǽʹץ֥Ȥο */
#define MAX_RECTANGLE		256	/* Ѳǽʶ֥Ȥο */

#define rotate_next(d)		(((d) + 1) % 4)
#define rotate_prev(d)		(((d) + 3) % 4)

/* ײˡ */
const dimension_t image_size[MAX_SURFACE][MAX_DIRECTION] = {
  { { 24, 42 }, { 20, 46 }, { 24, 42 }, { 20, 46 } }, /*  */
  { { 24, 42 }, { 32, 35 }, { 24, 42 }, { 32, 35 } }, /* ɽ */
  { { 24, 42 }, { 32, 35 }, { 24, 42 }, { 32, 35 } }  /*  */
};

/* ǥåˡ */
const rectangle_t layout_deck = { 0, 0, 576, 584 };

/* ΰ */
const rectangle_t layout_wall = { 204, 244, 168, 90 };

/* ΰ */
const rectangle_t layout_hand[MAX_DIRECTION] = {
  {  96, 504, 440,  64 },
  { 504,  16,  48, 480 },
  {  40,  16, 440,  64 },
  {  24,  88,  48, 480 }
};

/* פλ */
static const point_t hand_origin[MAX_SURFACE][MAX_DIRECTION] = {
  { {  96, 514 }, { 512, 474 }, { 480,  18 }, {  40,  88 } },
  { {  96, 504 }, { 504, 485 }, { 480,  38 }, {  40,  88 } },
  { {  96, 504 }, { 504, 485 }, { 480,  38 }, {  40,  88 } }
};

/* Ϫλ */
static const point_t exposure_origin[MAX_DIRECTION] = {
  { 536, 526 },
  { 520,  16 },
  {  40,  16 },
  {  24, 558 }
};

/* Τΰ */
const rectangle_t layout_bucket[MAX_DIRECTION] = {
  { 184, 384, 224, 112 },
  { 400, 136,  96, 240 },
  { 168,  88, 224, 112 },
  {  80, 208,  96, 240 }
};

/* Τΰλ */
static const point_t bucket_origin[MAX_DIRECTION] = {
  { 184, 384 },
  { 400, 365 },
  { 392, 158 },
  { 144, 208 }
};

/* Ωľΰ */
const rectangle_t layout_ready[MAX_DIRECTION] = {
  { 252, 368, 72, 7 },
  { 384, 252, 8, 72 },
  { 252, 212, 72, 7 },
  { 184, 252, 8, 72 }
};

/* ȥޡ */
const rectangle_t layout_yagamo[MAX_DIRECTION] = {
  { 136, 456, 40, 40 },
  { 456, 384, 40, 40 },
  { 400,  88, 40, 40 },
  {  80, 160, 40, 40 }
};

static const dimension_t face[MAX_DIRECTION] = {
  { 24, 32 }, { 32, 24 }, { 24, 32 }, { 32, 24 }
};

static const vector_t step[MAX_DIRECTION] = {
  { 24, 32 }, { 32, -24 }, { -24, -32 }, { -32, 24 }
};

static rectangle_t *new_rectangle(int x, int y, int width, int height)
{
  static rectangle_t buffer[MAX_RECTANGLE];
  static int avail[MAX_RECTANGLE], size = 0;
  rectangle_t *top;

  if (size == 0) {
    /* ߽ */
    int marked[MAX_RECTANGLE];
    int i;
    for (i = 0; i < MAX_RECTANGLE; i++) {
      marked[i] = 0;
    }
    for (i = 0; i < MAX_PLAYER; i++) {
      player_t *p = &players[i];
      int n;
      for (n = 0; n < MAX_SORT; n++) {
        if (p->discard[n] != NULL) marked[p->discard[n] - buffer] = 1;
        if (p->through[n] != NULL) marked[p->through[n] - buffer] = 1;
      }
    }
    for (i = 0; i < MAX_RECTANGLE; i++) {
      if (!marked[i])
        avail[size++] = i;
    }
    if (size == 0) {
      fprintf(stderr, "new_rectangle(): Out of memory.\n");
      exit(1);
    }
  }
  top = &buffer[avail[--size]];
  top->x = x;
  top->y = y;
  top->width = width;
  top->height = height;
  return top;
}

bone_t *chain_bone(int tid, int x, int y, bone_t *next)
{
  static bone_t buffer[MAX_BONE], *end = NULL;
  bone_t *top;
  
  if (end == NULL) {
    /* ߽ */
    int marked[MAX_BONE];
    int i;
    bone_t *b;

    for (i = 0; i < MAX_BONE; i++) {
      marked[i] = 0;
    }
    for (b = next; b != NULL; b = b->next) {
      marked[b - buffer] = 1;
    }
    for (b = wall_bones; b != NULL; b = b->next) {
      marked[b - buffer] = 1;
    }
    for (i = 0; i < MAX_PLAYER; i++) {
      for (b = players[i].bucket;   b != NULL; b = b->next)
        marked[b - buffer] = 1;
      for (b = players[i].exposure; b != NULL; b = b->next)
        marked[b - buffer] = 1;
    }
    for (i = 0; i < MAX_BONE; i++) {
      if (!marked[i])
        buffer[i].next = end, end = &buffer[i];
    }
    if (end == NULL) {
      fprintf(stderr, "chain_exposure(): Out of memory.\n");
      exit(1);
    }
  }
  top = end; end = end->next;
  top->tid = tid;
  top->x = x;
  top->y = y;
  top->next = next;
  return top;
}

rectangle_t *bucket_append(int pid, int n, int angled)
{
  int direction = pid;
  int column = 0;
  bone_t *b;  
  int tid, x, y;

  if (angled) {
    tid = make_tid(n, rotate_next(direction), SURFACE_FACE);
    tid = toggle_angled(tid);
  } else {
    tid = make_tid(n, direction, SURFACE_FACE);
  }
  x = bucket_origin[direction].x;
  y = bucket_origin[direction].y;
  
  if (players[pid].bucket != NULL) {
    int count = 0;
    for (b = players[pid].bucket; b != NULL; b = b->next) {
      count++;
    }
    
    column = count % WRAP_COUNT;

    if (direction % 2 == 0) {
      if (column != 0) {
        x = players[pid].bucket->x;
      }
      y += step[direction].y * (count / WRAP_COUNT);
      if (angled)
        y += 4;
    } else {
      if (column != 0) {
        y = players[pid].bucket->y;
      }
      x += step[direction].x * (count / WRAP_COUNT);      
      if (angled)
        x += 4;
    }
  }
  
  b = players[pid].bucket;
  switch (direction) {
  case 0:
    if (b != NULL && column != 0) {
      if (is_angled(b->tid))
        x += face[1].width;
      else
        x += face[0].width;
    }
    break;
  case 1:
    if (angled)
      y -= face[0].height;
    else
      y -= face[1].height;
    break;
  case 2:
    if (angled)
      x -= face[1].width;
    else
      x -= face[0].width;
    break;
  case 3:
    if (b != NULL && column != 0) {
      if (is_angled(b->tid))
        y += face[0].height;
      else
        y += face[1].height;
    }
    break;
  }
 
  players[pid].bucket = chain_bone(tid, x, y, players[pid].bucket);
  
  if (angled) {
    int d = rotate_next(direction);
    return new_rectangle(x, y, face[d].width, face[d].height);
  } else
    return new_rectangle(x, y, face[direction].width, face[direction].height);
}

static rectangle_t *exposure_append_tile(int pid, int n, int angled, int surface)
{
  int direction = pid;
  int tid, x, y;
  bone_t *b;

  if (angled) {
    tid = make_tid(n, rotate_next(direction), surface);
    tid = toggle_angled(tid);
  } else {
    tid = make_tid(n, direction, surface);
  }
  
  x = exposure_origin[direction].x;
  y = exposure_origin[direction].y;

  b = players[pid].exposure;
  if (b != NULL) {
    if (direction % 2 == 0)
      x = b->x;
    else
      y = b->y;
  }
  switch (direction) {
  case 0:
    if (angled) {
      x -= face[1].width;
      y += 7;
    } else
      x -= face[0].width;
    break;
  case 1:
    if (angled)
      x += 8;
    if (b != NULL) {
      if (is_angled(b->tid))
        y += face[0].height;
      else
        y += face[1].height;
    }
    break;
  case 2:
    if (b != NULL) {
      if (is_angled(b->tid))
        x += face[1].width;
      else
        x += face[0].width;
    }
    break;
  case 3:
    if (angled)
      y -= face[0].height;
    else
      y -= face[1].height;
    break;
  }
  players[pid].exposure = chain_bone(tid, x, y, players[pid].exposure);
  
  if (angled) {
    int d = rotate_next(direction);
    return new_rectangle(x, y, face[d].width, face[d].height);
  } else
    return NULL;
}

rectangle_t *exposure_append(int pid, set_t set)
{
  int result = 0;
  int n = tile_of(set);
  int m = mark_of(set);
  switch (type_of(set)) {
  case SEQUENCE:
    switch (m) {
    case 1:
      exposure_append_tile(pid, n + 1, 0, SURFACE_FACE);
      exposure_append_tile(pid, n + 2, 0, SURFACE_FACE);
      result = (int)exposure_append_tile(pid, n, 1, SURFACE_FACE);
      break;
    case 2:
      exposure_append_tile(pid, n,     0, SURFACE_FACE);
      exposure_append_tile(pid, n + 2, 0, SURFACE_FACE);
      result = (int)exposure_append_tile(pid, n + 1, 1, SURFACE_FACE);      
      break;
    case 3:
      exposure_append_tile(pid, n,     0, SURFACE_FACE);
      exposure_append_tile(pid, n + 1, 0, SURFACE_FACE);
      result = (int)exposure_append_tile(pid, n + 2, 1, SURFACE_FACE);
      break;
    }
    break;
  case TRIPLET:
    result |= (int)exposure_append_tile(pid, n, m == 1, SURFACE_FACE);
    result |= (int)exposure_append_tile(pid, n, m == 2, SURFACE_FACE);
    result |= (int)exposure_append_tile(pid, n, m == 3, SURFACE_FACE);
    break;
  case QUADRUPLET:
    result |= (int)exposure_append_tile(pid, n, m == 1, SURFACE_FACE);
    result |= (int)exposure_append_tile(pid, n, 0,      SURFACE_FACE);
    result |= (int)exposure_append_tile(pid, n, m == 2, SURFACE_FACE);
    result |= (int)exposure_append_tile(pid, n, m == 3, SURFACE_FACE);
    break;
  }
  return (rectangle_t *)result;
}

rectangle_t *exposure_append_quadruplet(int pid, int n)
{
  exposure_append_tile(pid, n, 0, SURFACE_BACK);
  exposure_append_tile(pid, n, 0, SURFACE_FACE);
  exposure_append_tile(pid, n, 0, SURFACE_FACE);
  exposure_append_tile(pid, n, 0, SURFACE_BACK);
  return NULL;
}

rectangle_t *exposure_append_quadruplet_lifting(int pid, int n)
{
  int direction = pid;
  int d = rotate_next(direction);
  int tid;
  bone_t *b;
  
  tid = make_tid(n, d, SURFACE_FACE);
  tid = toggle_angled(tid);
  
  for (b = players[pid].exposure; b != NULL; b = b->next) {
    if (b->tid == tid) {
      int x, y;
      if (direction % 2 == 0) {
        x = b->x;
        y = b->y + step[d].y;
      } else {
        x = b->x + step[d].x;
        y = b->y;
      }
      b->next = chain_bone(b->tid, b->x, b->y, b->next);
      b->x = x;
      b->y = y;
      return new_rectangle(x, y, face[d].width, face[d].height);
    }
  }
  fprintf(stderr, "exposure_append_quadruplet_lifting(): Not found.\n");
  exit(1);
  return NULL;
}

void align_handrail(int pid, int n, int surface)
{
  int *hand = players[pid].hand;
  handrail_t *h = &players[pid].handrail;
  int direction = pid;
  int x, y;
  int i, j, tiles[14];
  bone_t *end;

  h->count = 0;

  if (n >= 0) {
    hand[n]--; /* פ */
  }
  for (i = 0; i < MAX_SORT; i++) {
    for (j = hand[i]; j > 0; j--)
      tiles[h->count++] = i;
  }
  if (n >= 0) {
    tiles[h->count++] = n;
    hand[n]++; /* ᤹ */
  }

  x = hand_origin[surface][direction].x;
  y = hand_origin[surface][direction].y;

  end = h->__buf;
  end->next = NULL;

  switch (direction) {
  case 0:
    for (i = 0; i < h->count; i++) {
      end->tid = make_tid(tiles[i], direction, surface);
      end->x = x;
      end->y = y;
      end++;
      end->next = end - 1;
      x += face[0].width;
    }
    break;
  case 1:
    for (i = 0; i < h->count; i++) {
      end->tid = make_tid(tiles[i], direction, surface);
      y -= face[1].height;
      end->x = x;
      end->y = y;
      end++;
      end->next = end - 1;
    }
    break;
  case 2:
    for (i = 0; i < h->count; i++) {
      end->tid = make_tid(tiles[i], direction, surface);
      x -= face[0].width;
      end->x = x;
      end->y = y;
      end++;
      end->next = end - 1;
    }
    break;
  case 3:
    for (i = 0; i < h->count; i++) {
      end->tid = make_tid(tiles[i], direction, surface);
      end->x = x;
      end->y = y;
      end++;
      end->next = end - 1;
      y += face[1].height;
    }
    break;
  }
 
  h->bones = end->next;
}

void toggle_handrail_surface(int pid, bone_t *b, int surface)
{
  int direction = pid;
  
  b->tid = make_tid(tile_of(b->tid), direction, surface);
  if (direction % 2 == 0)
    b->y = hand_origin[surface][direction].y;
  else
    b->x = hand_origin[surface][direction].x;
}

/*  */
bone_t *make_wall_bones(int *tomb)
{
  bone_t *b = NULL;
  int x = layout_wall.x;
  int y = layout_wall.y;
  int tid;
  int i;

  for (i = 0; i < 14; i += 2) {
    tid = make_tid(tomb[i + 1], 0, SURFACE_BACK);
    b = chain_bone(tid, x, y + 15, b);

    tid = make_tid(tomb[i], 0, SURFACE_BACK);
    b = chain_bone(tid, x, y, b);
    x += face[0].width;
  }
  return b;
}
