/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: detect.c,v 1.1.1.1 1999/11/05 01:55:18 shinmaru Exp $
 */

#include "mahjong.h"
#include "option.h"

#define hook_append(h, t)	((h).needs++, (h).types |= bit(t))

static int __min_going;
static hook_table_t __h;

static void receiver(set_t *sets, int count)
{
  int going = calculate_going(sets, count);
  if (going > __min_going) return;
  if (going < __min_going) {
    __min_going = going;
    memset(__h, 0, sizeof(__h));
  }
  if (!option_silence) {
    int i;
    for (i = 0; i < count; i++) {
      print_set(sets[i]);
    }
    printf(": %d\n", going);
  }
  detect_hooks(sets, count, going, __h);
}

int calculate_hand(int *hand, hook_table_t h)
{
  int count, types;
  int i;

  __min_going = INT_MAX;

  /* ̷ */
  generate_combinations(hand, receiver);
  memcpy(h, __h, sizeof(hook_table_t));
  
  for (i = 0, count = 0, types = 0; i < MAX_SORT; i++) {
    count += hand[i];
    /* פǻȤڤäƤפǤԤ̵ˤ */
    if (hand[i] == 4)
      h[i].types = 0;
    else
      types |= h[i].types;
  }
  if (types == 0)
    __min_going += 1;
  
  if (count >= 13) {
    int going;
    
    /* л */
    memset(__h, 0, sizeof(__h));    
    going = calculate_going_7_pairs(hand, __h);
    if (going == __min_going) {
      for (i = 0; i < MAX_SORT; i++)
        h[i].types |= __h[i].types;
    } else if (going < __min_going) {
      __min_going = going;
      memcpy(h, __h, sizeof(hook_table_t));
    }
    
    /* ̵ */
    memset(__h, 0, sizeof(__h));
    going = calculate_going_13_orphans(hand, __h);
    if (going == __min_going) {
      for (i = 0; i < MAX_SORT; i++)
        h[i].types |= __h[i].types;
    } else if (going < __min_going) {
      __min_going = going;
      memcpy(h, __h, sizeof(hook_table_t));
    }
  }
  return __min_going;
}

void detect_hooks(set_t *sets, int count, int going, hook_table_t h)
{
  int pairs = 0;
  int i;

  for (i = 0; i < count; i++) {
    if (type_of(sets[i]) == PAIR)
      pairs++;
  }
  for (i = 0; i < count; i++) {
    set_t set = sets[i];
    int n = tile_of(set);
    if (is_revealed(set))
      continue;
    switch (type_of(set)) {
    case SINGLETON:
      if (pairs == 0 && going == 1) hook_append(h[n], HOOK_PAIR);
      break;
    case PAIR:
      if (pairs >= 2) hook_append(h[n], HOOK_TRIPLET);
      break;
    case CONJUNCT:
      if (pairs == 0 && going == 1) {
        /* ñ */
        hook_append(h[n    ], HOOK_PAIR);
        hook_append(h[n + 1], HOOK_PAIR);
      } else {
        if (number_of(n) > 0) hook_append(h[n - 1], HOOK_BELOW);
        if (number_of(n) < 7) hook_append(h[n + 2], HOOK_ABOVE);
      }
      break;
    case SPLITTER:
      if (pairs == 0 && going == 1) {
        /* ñ */
        hook_append(h[n    ], HOOK_PAIR);
        hook_append(h[n + 2], HOOK_PAIR);
      } else
        hook_append(h[n + 1], HOOK_MIDDLE);
      break;
    }
  }
}

int find_hook_types(set_t *sets, int count, int last_tile)
{
  int hook_types = 0;
  int i;
  for (i = 0; i < count; i++) {
    set_t set = sets[i];
    if (!is_revealed(set)) {
      int n = tile_of(set);
      switch (type_of(set)) {
      case SINGLETON:
        if (last_tile == n)
          hook_types |= bit(HOOK_SINGLETON);
        break;
      case PAIR:
        if (last_tile == n)
          hook_types |= bit(HOOK_PAIR);
        break;
      case SEQUENCE:
        if (last_tile == n)
          hook_types |= bit(HOOK_BELOW);
        else
        if (last_tile == n + 1)
          hook_types |= bit(HOOK_MIDDLE);
        else
        if (last_tile == n + 2)
          hook_types |= bit(HOOK_ABOVE);
        break;
      case TRIPLET:
        if (last_tile == n)
          hook_types |= bit(HOOK_TRIPLET);
        break;
      case QUADRUPLET:
        if (last_tile == n)
          hook_types |= bit(HOOK_QUADRUPLET);
        break;
      }
    }
  }
  return hook_types;
}

int calculate_going_7_pairs(int *hand, hook_table_t h)
{
  int pairs = 0, triplets = 0, going;
  int i;
  for (i = 0; i < MAX_SORT; i++) {
    if (hand[i] >= 2) {
      if (hand[i] > 2) triplets++;
      pairs++;
    }
  }
  going = 7 - pairs;
  if (going == 1) {
    if (triplets >= 2) {
      return going + 1; /* [3/3/2/2/2/2]η */
    }
    for (i = 0; i < MAX_SORT; i++) {
      if (hand[i] == 1)
        hook_append(h[i], HOOK_PAIR);
    }
  }
  return going;
}

int calculate_going_13_orphans(int *hand, hook_table_t h)
{
  int pairs = 0, going = 0;
  int i;
  
  for (i = 0; i < 13; i++) {
    int n = hand[terminal_tiles[i]];
    if (n > 0) {
      going++;      
      if (n >= 2) pairs++;
    }
  }
  going = 14 - going;
  if (pairs >= 1)
    going--;

  if (going == 1) {
    if (pairs == 0) {
      /* 13Ԥ */        
      for (i = 0; i < 13; i++)
        hook_append(h[terminal_tiles[i]], HOOK_PAIR);
    } else {
      for (i = 0; i < 13; i++) {
        int n = terminal_tiles[i];
        if (hand[n] == 0) {
          hook_append(h[n], HOOK_SINGLETON);
          break;
        }
      }
    }
  }
  return going;
}
