/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: generate.c,v 1.1.1.1 1999/11/05 01:55:18 shinmaru Exp $
 */

#include "mahjong.h"

static set_t result[MAX_HANDED_SET];
static int null_ignore[MAX_SORT];

static int *subtract(int *, set_t);
static int *filter(int *, set_t);
static int parse(int *, int);
static void loop(int *hand, int *ignore, int count, receiver_t receiver);

void generate_combinations(int *hand, receiver_t receiver)
{
  loop(hand, null_ignore, 0, receiver);
}

void loop(int *hand, int *ignore, int count, receiver_t receiver)
{
  int n;
  for (n = 0; n < MAX_SORT && hand[n] == 0; n++);
  if (n < MAX_SORT) {
    int bits = parse(hand, n) & ~ignore[n];
    int type;
    /* Ĵ٤İȹ礻󤷤ˤǤ */
    for (type = MAX_TYPEOF_SET - 1; type >= 0; type--) {
      if (bittest(bits, type)) {
        int hand_next[MAX_SORT], ignore_next[MAX_SORT];
        int i;
        set_t set;
        for (i = 0; i < MAX_SORT; i++) {
          hand_next[i] = hand[i]; ignore_next[i] = ignore[i];
        }
        set = make_set(n, type);
        result[count] = set;
        loop(subtract(hand_next, set), filter(ignore_next, set), count + 1, receiver);
      }
    }
  } else
    receiver(result, count);
}

int parse(int *hand, int n)
{
  int bits = bit(SINGLETON);
  if (is_numeric(n)) {
    int n1 = hand[n];
    int n2 = number_of(n) < 8 ? hand[n + 1] : 0;
    int n3 = number_of(n) < 7 ? hand[n + 2] : 0;
    if (n1 && n2) bits |= bit(CONJUNCT);
    if (n1 && n3) bits |= bit(SPLITTER);
    if (n1 && n2 && n3) bits |= bit(SEQUENCE);
  }
  switch (hand[n]) {
  case 2: bits |= bit(PAIR); break;
  case 3: bits |= bit(TRIPLET); break;
  case 4: break;
  }
  return bits;
}

int *subtract(int *hand, set_t set)
{
  int n = tile_of(set);
  switch (type_of(set)) {
  case SINGLETON:  hand[n] -= 1; break;
  case PAIR:       hand[n] -= 2; break;
  case CONJUNCT:   hand[n]--; hand[n + 1]--; break;
  case SPLITTER:   hand[n]--; hand[n + 2]--; break;
  case SEQUENCE:   hand[n]--; hand[n + 1]--; hand[n + 2]--; break;
  case TRIPLET:    hand[n] -= 3; break;
  case QUADRUPLET: hand[n] -= 4; break;
  }
  return hand;
}

int *filter(int *ignore, set_t set)
{
  int n = tile_of(set);
  switch (type_of(set)) {
  case SINGLETON:
    ignore[n] |= bit(SINGLETON) | bit(CONJUNCT) | bit(SPLITTER) | bit(SEQUENCE);
    if (is_numeric(n)) {
      if (number_of(n) < 8)
        ignore[n + 1] |= bit(SINGLETON) | bit(CONJUNCT);
      if (number_of(n) < 7)
        ignore[n + 2] |= bit(SINGLETON);
    }
    break;
  case CONJUNCT:
    ignore[n] |= bit(SPLITTER);
    if (number_of(n) < 7) /* פ˷ޤäƤ */
      ignore[n + 2] |= bit(SINGLETON);
    break;
  case SPLITTER:
    ignore[n] |= bit(CONJUNCT);
    ignore[n + 1] |= bit(SINGLETON) | bit(CONJUNCT) | bit(SPLITTER);
    break;
  }
  return ignore;
}
