/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: hand_main.c,v 1.2 1999/11/06 01:07:49 shinmaru Exp $
 */

#include "mahjong.h"
#include "debug.h"
#include "option.h"

#include <ctype.h>

int main(int argc, char *argv[])
{
  const int pid = 0;
  char buffer[BUFSIZ];
  int hand[MAX_SORT];

  if (argc == 2 && strcmp(argv[1], "-s") == 0) {
    option_silence = 1;
  }
  fflush(stdin);

  roundkeeper.rounds = 2;
  
  while (option_silence ? 0 : fprintf(stderr, "\nhand>"),
         fgets(buffer, sizeof(buffer), stdin) != NULL) {
    int suit_base = 0;
    int i, j, count;
    int last_tile;
    const char *p;

    for (i = 0; i < MAX_SORT; i++) hand[i] = 0;
    for (p = buffer; *p != '\0'; p++) {
      if (isspace(*p)) continue;
      if (*p == '/') {
        suit_base += 9; if (suit_base > 18) suit_base = 0;
      } else {
        switch (toupper(*p)) {
        case '1': case '2': case '3':
        case '4': case '5': case '6':
        case '7': case '8': case '9':
          last_tile = *p - '1' + suit_base; break;
        case 'H': last_tile = 27; break;
        case 'G': last_tile = 28; break;
        case 'R': last_tile = 29; break;
        case 'E': last_tile = 30; break;
        case 'S': last_tile = 31; break;
        case 'W': last_tile = 32; break;
        case 'N': last_tile = 33; break;
        default:
          fprintf(stderr, "Unrecognized character: %c\n", *p);
          goto error;
        }
        hand[last_tile]++;
      }
    }

    count = 0;
    for (i = 0; i < MAX_SORT; i++) {
      count += hand[i];      
      if (hand[i] > 4) {
        fprintf(stderr, "Too many tiles of: %s\n", tile_face[i]);
        goto error;
      }
    }
    if (count > 14) {
      fprintf(stderr, "Too many tiles in hand:");
      for (i = 0; i < MAX_SORT; i++) {
        for (j = hand[i]; j > 0; j--) {
          fprintf(stderr, "%s", tile_face[i]);
        }
      }
      fprintf(stderr, "\n");
      goto error;
    }
    if (count % 3 == 2) {
      for (i = 0; i < MAX_SORT; i++) {
        players[pid].hand[i] = hand[i];
      }
      evaluate(pid, last_tile);
      if (count == 14 && players[pid].scoresheet != NULL) {
        print_scoresheet(players[pid].scoresheet);
      } else {
        print_going_table();
      }
    }
    else if (count != 0) {
      hook_table_t h;
      int going;
      
      going = calculate_hand(hand, h);
      if (option_silence) {
        for (i = 0; i < MAX_SORT; i++) {
          if (h[i].types != 0 && hand[i] < 4)
            printf("%s", tile_face[i]);
        }
      } else if (going == 0) {
        printf("λ");
      } else {
        printf(going > 1 ? "%dİ:" : "İ:", going - 1);
        print_hook_table(h);
      }
      putchar('\n');
    }
  error:
  }
  return 0;
}
