/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: interactive.c,v 1.3 1999/11/10 04:54:36 shinmaru Exp $
 */

#include "mahjong.h"
#include "graphics.h"

#define interactive_situation() (__situation == CHOOSE || __situation == NOTIFIED)

#define CHOOSE		1
#define NOTIFIED	2

#define WIDTH		(image_size[0][0].width)
#define HEIGHT		(image_size[0][0].height)

#define resume(e)	(reset_highlight(), reset_selection(), __situation = 0, dispatch(e))

#define in_rectangle(x, y, r)				\
  ((r)->x <= (x) && (x) < (r)->x + (r)->width &&	\
   (r)->y <= (y) && (y) < (r)->y + (r)->height)

static int __forbids_quadruplet;
static int __interest;
static int __notified;
static int __situation;

extern void show_popup_menu(int x, int y, int winning, int quadruplet);

static int interesting_types(int *hand, int n)
{
  int types = 0;
  switch (hand[n]) {
  case 3: types |= bit(HOOK_QUADRUPLET); /* ȴ */    
  case 2: types |= bit(HOOK_TRIPLET);
  }
  if (is_numeric(n)) {
    int m2 = number_of(n) >= 2 ? hand[n - 2] : 0;
    int m1 = number_of(n) >= 1 ? hand[n - 1] : 0;
    int n1 = number_of(n) <= 7 ? hand[n + 1] : 0;
    int n2 = number_of(n) <= 6 ? hand[n + 2] : 0;
    if (n2 && n1) types |= bit(HOOK_BELOW);
    if (m2 && m1) types |= bit(HOOK_ABOVE);
    if (m1 && n1) types |= bit(HOOK_MIDDLE);
  }
  return types;
}

static int expose_toggled_bones(int *toggled)
{
  int types = interesting_types(toggled, __notified) & __interest;

  if (toggled[__notified] == 3) {
    types &= ~bit(HOOK_TRIPLET);
  }
  if (types == bit(HOOK_TRIPLET) ||
      types == bit(HOOK_QUADRUPLET) ||
      types == bit(HOOK_BELOW) ||
      types == bit(HOOK_ABOVE) ||
      types == bit(HOOK_MIDDLE)) {
    /* ǤξƤȤ */
    players[0].robbings[__notified].types = types;
    return 1;
  } else
    return 0;
}

static int expose_toggled_quadruplet(int *toggled, int count)
{
  int i;

  if (__forbids_quadruplet)
    return -1;

  if (count == 4) {
    for (i = 0; i < MAX_SORT; i++) {
      if (toggled[i] == 4)
	return i;
    }
  }
  else if (count == 1) {
    for (i = 0; i < players[0].revealed_count; i++) {
      set_t set = players[0].revealed[i];
      int n = tile_of(set);
      if (type_of(set) == TRIPLET && toggled[n] == 1)
	return n;
    }
  }
  return -1;
}

static bone_t *handrail_hittest(int x, int y)
{
  bone_t *b;
  for (b = players[0].handrail.bones; b != NULL; b = b->next) {
    if (b->x <= x && x < b->x + WIDTH &&
        b->y <= y && y < b->y + HEIGHT)
      break;
  }
  return b;
}

int interactive_choose(int pid, int n, int forbids_quadruplet)
{
  __forbids_quadruplet = forbids_quadruplet;
  __notified = n;
  __situation = CHOOSE;

  players[0].going = evaluate(pid, n);
  show_selection();

  return ACTION_SUSPEND;
}

int interactive_notified(int pid, int n, int permission)
{
  player_t *p = &players[pid];
  p->robbings[n].types = 0;

  __interest = interesting_types(p->hand, n) & permission;

  if (p->winnings[n].types != 0 || (!is_ready(pid) && __interest != 0)) {
    __situation = NOTIFIED;
    __notified = n;

    set_highlight(players[current_discarder()].discard[n]);

    return ACTION_SUSPEND;
  } else
    return REPLY_NEXT;
}

void pointer_move(int x, int y)
{
#ifdef __WIN32__
  if (__situation == CHOOSE) {
    bone_t *b;

    reset_selection();
    
    b = handrail_hittest(x, y);
    if (b != NULL) {
      static rectangle_t pointed;

      pointed.x = b->x;
      pointed.y = b->y + (surface_of(b->tid) == SURFACE_HEAD ? 10 : 0);
      pointed.width = WIDTH;
      pointed.height = 32;
      set_selection(&pointed);
    }
  }
#endif
}

void button1_down(int x, int y)
{
  player_t *p = &players[0];
  int total = 0, number_of_toggled = 0;
  bone_t *b;
  int i, toggled[MAX_SORT];

  for (i = 0; i < MAX_SORT; i++)
    toggled[i] = 0;

  for (b = p->handrail.bones; b != NULL; b = b->next, total++) {
    if (surface_of(b->tid) == SURFACE_FACE)
      toggled[tile_of(b->tid)]++, number_of_toggled++;
  }

  switch (__situation) {
  case CHOOSE:
    if (number_of_toggled == total && __notified >= 0) {
      /*  */
      resume(event_winning(0, __notified));
    }
    else if (number_of_toggled == 0) {
      if ((b = handrail_hittest(x, y)) != NULL) {
	int n = tile_of(b->tid);

	if (!is_ready(0)) {
	  /*  */
	  p->going = refer_going(n);

	  if (p->going == 1)
	    memcpy(p->winnings, going_table[n].hook, sizeof(hook_table_t));
	  else
	    memset(p->winnings, 0, sizeof(hook_table_t));

	  resume(event_discard(0, n));
	} else {
	  /* ڤ */
	  resume(event_discard(0, __notified));
	}
      }
    } else {
      int n = expose_toggled_quadruplet(toggled, number_of_toggled);
      if (n >= 0) {
	if (is_ready(0) && 
	    memcmp(p->winnings, going_table[n].hook, sizeof(hook_table_t)) != 0) {
	  /* ԤѤ륫Ͻʤ */
	  beep();
	  return;
	}
	/*  */
	resume(event_quadruple(0, n));
      }
    }
    break;

  case NOTIFIED:
    if (number_of_toggled == total) {
      /*  */
      p->hand[__notified]++;
      evaluate(0, __notified);
      p->hand[__notified]--;
      resume(reply_winning(0, __notified));
    }
    else if ((b = handrail_hittest(x, y)) != NULL) {
      if (expose_toggled_bones(toggled)) {
	resume(REPLY_NEXT);
      }
    }
    else if (number_of_toggled == 0) {
      /* פƤʤʤСƨ */
      resume(REPLY_NEXT);
    }
    break;
  }
}

void button2_down(int x, int y)
{
  player_t *p = &players[0];
  bone_t *b;

  if (__situation != CHOOSE ||
      is_ready(0) ||
      bittest(p->condition, COND_REVEALED))
    return;

  if ((b = handrail_hittest(x, y)) != NULL) {
    int n = tile_of(b->tid);
    int going = refer_going(n);
    if (going != 1) {
      beep();
      return;
    } else {
      /* Ωľ */
      memcpy(p->winnings, going_table[n].hook, sizeof(hook_table_t));
      resume(event_ready(0, n));
    }
  }
}

void button3_down(int x, int y)
{
  bone_t *b;

  if (!interactive_situation())
    return;

  if ((b = handrail_hittest(x, y)) != NULL) {
    clear_rectangle(b->x, b->y, WIDTH, HEIGHT);
    toggle_handrail_surface(0, b, (surface_of(b->tid) != SURFACE_HEAD
				   ? SURFACE_HEAD
				   : SURFACE_FACE));
    put_tile(b->x, b->y, b->tid);
    pointer_move(x, y);
  } else {
    int total = 0, number_of_toggled = 0;

    for (b = players[0].handrail.bones; b != NULL; b = b->next, total++) {
      if (surface_of(b->tid) == SURFACE_FACE)
	number_of_toggled++;
    }
    
    if (number_of_toggled == 0) {
      /*  */
      for (b = players[0].handrail.bones; b != NULL; b = b->next) {
	toggle_handrail_surface(0, b, SURFACE_FACE);
      }
    }
    else {
      /* פ򸵤᤹ */
      for (b = players[0].handrail.bones; b != NULL; b = b->next) {
	toggle_handrail_surface(0, b, SURFACE_HEAD);
      }
    }
    repaint_handrail(0);
  }
}

int interactive_report_winning(const char *reason, int pid, int n, int *bill)
{
  char text[1024], *p = text;
  int self_drawn;
  scoresheet_t *sheet = players[pid].scoresheet;

  self_drawn = pid == current_discarder();

  if (!self_drawn) {
    set_highlight(players[current_discarder()].discard[n]);

    if (sheet != NULL) {
      rectangle_t *screening = NULL;
      int alt;

      for (alt = 0; alt < MAX_SORT; alt++) {
	if (players[pid].winnings[alt].types != 0) {
	  screening = players[pid].discard[alt];
	  if (screening != NULL)
	    break;
	  screening = players[pid].through[alt];
	  if (screening != NULL)
	    break;
	}
      }
      if (screening != NULL) {
	/* İ */
	static char buf[256];

	sprintf(buf, "%s(İ)", reason);
	reason = buf;
	set_screening(screening);
      }
    }
  }
  p += sprintf(p, "\n%s\n", reason);
  if (sheet != NULL) {
    char buf[256], *t = buf;
    hand_name_t *hand;

    t += sprintf(t, sheet->fmt, sheet->point, sheet->bonus);
    sprintf(t, " %d", sheet->total);

    p += sprintf(p, "%30s\n\n", buf);
    for (hand = sheet->names; hand != NULL; hand = hand->next) {
      p += sprintf(p, hand->bonus > 0 ? "%-27s%d\n" : "%s\n",
                   hand->name, hand->bonus);
    }
  }

  report_dialog(text);
  score_dialog(bill);

  if (!self_drawn) {
    reset_highlight();
    reset_screening();
  }
  return REPORT_OK;
}

int interactive_report_round_over(const char *reason, int *bill)
{
  report_dialog(reason);
  score_dialog(bill);

  reset_highlight();
  return REPORT_OK;
}
