/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: mahjong.h,v 1.3 1999/11/10 04:54:36 shinmaru Exp $
 */

#ifndef mahjong_H
#define mahjong_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <time.h>

#ifdef __BORLANDC__
#include <mem.h>
#endif

#ifdef __GNUC__
#define max(a, b)		((a) > (b) ? (a) : (b))
#define min(a, b)		((a) < (b) ? (a) : (b))
#endif

#define bittest(u, n)		(((u) & bit(n)) != 0u)
#define bit(n)			(1u << (n))

#ifdef __BORLANDC__
#define srandom(s)		(srand(s))
#define random_integer(n)	((rand() >> 8) % (n))
#else
#define random_integer(n)	((random() >> 8) % (n))
#endif

typedef struct {
  int		x;
  int		y;
  int		width;
  int		height;
} rectangle_t;

typedef struct {
  int x;
  int y;
} point_t;

typedef point_t vector_t;

typedef struct {
  int width;
  int height;
} dimension_t;

/*
 * 
 */

#define MAX_SORT		34	/* ׼ */
#define MAX_SUIT		4	/* ʬ */

#define SUIT_CHARACTER		0	/* ߻ */
#define SUIT_BALL		1	/*  */
#define SUIT_BAMBOO		2	/*  */
#define SUIT_HONOR		3	/*  */

#define is_character(n)		             ((n) <  9)
#define is_ball(n)		( 9 <= (n) && (n) < 18)
#define is_bamboo(n)		(18 <= (n) && (n) < 27)
#define is_dragon(n)		(27 <= (n) && (n) < 30)
#define is_wind(n)		(30 <= (n) && (n) < 34)

#define number_of(n)		((n) % 9)
#define suit_of(n)		((n) / 9)

#define is_numeric(n)		((n) <  27)
#define is_honor(n)		((n) >= 27)

extern const char *tile_face[MAX_SORT]; /* ײ */

extern const int terminal_tiles[13];	/* ϷƬ */

/*
 * ̻
 */

typedef int set_t;

#define MAX_HANDED_SET		16	/* Ȥ߹礻뤳ȤǤ̻Ҥο */
#define MAX_TYPEOF_SET		7	/* ̻Ҥμ */

#define SINGLETON		0	/* Ω */
#define PAIR			1	/* л */
#define CONJUNCT		2	/* ξ */
#define SPLITTER		3	/*  */
#define SEQUENCE		4	/*  */
#define TRIPLET			5	/*  */
#define QUADRUPLET		6	/* ʻ */

#define MARKING_RIGHT		1	/*  */
#define MARKING_FRONT		2	/*  */
#define MARKING_LEFT		3	/*  */

#define make_set(n, t)		((n) | (t) << 16)
#define mark_add(m, s)		((s) | (m) << 8 | 0x8000)

#define tile_of(s)		((s) & 0xff)
#define type_of(s)		((s) >> 16)
#define mark_of(s)		((s) >> 8 & 0x0f)

#define is_concealed(s)		(((s) & 0x8000) == 0)
#define is_revealed(s)		(((s) & 0x8000) != 0)
#define toggle_revealed_bit(s)	((s) | 0x8000)

#define tile_count(s)		(tile_count_of_set[type_of(s)])

extern int tile_count_of_set[MAX_TYPEOF_SET]; /* ̻ҤȤפθĿ */

extern int is_terminal(set_t set);
extern int point_of(set_t set, int pid);

extern void print_set(set_t set);

/*
 * Ԥ
 */

#define MAX_TYPEOF_HOOK		7	/* Ԥμ */

#define HOOK_SINGLETON		0	/* Ω */
#define HOOK_PAIR		1	/* л */
#define HOOK_TRIPLET		2	/*  */
#define HOOK_QUADRUPLET		3	/* ʻ */
#define HOOK_BELOW		4	/* () */
#define HOOK_ABOVE		5	/* () */
#define HOOK_MIDDLE		6	/* () */

/*
 * 
 */

#define MAX_EVAL		10	/* ο */

#define EVAL_NO_TERMINALS	0	/* 䥪 */
#define EVAL_TERMINALS		1	/*  */
#define EVAL_SEQUENCES		2	/* ʿ */
#define EVAL_SISTERS		3	/* ָ */
#define EVAL_STRAIGHT		4	/* 쵤̴ */
#define EVAL_3_SUIT_EQUALS	5	/* Ʊ */
#define EVAL_1_SUIT		6	/* 쿧 */
#define EVAL_TRIPLETS		7	/* С */
#define EVAL_7_PAIRS		8	/* л */
#define EVAL_13_ORPHANS		9	/* ̵ */

extern const int max_eval_values[MAX_EVAL];

typedef struct {
  int pid;				/* ץ쥤䡼 */
  set_t *sets;				/* ̻Ҥ¤ */
  int count;				/* ̻Ҥο */
  int going;				/* İ */
  int revealed;				/*  */
  int one_suit;				/* 쿧μ(̩˰) */
  int forbids;				/* Ω */
  int singletons[MAX_SORT], sequences[MAX_SORT], triplets[MAX_SORT];
} evaluated_hand_t;

/* ͭ */
typedef struct {
  char needs;				/* ɬפο */
  char types;				/* Ԥμ */
} hook_table_t[MAX_SORT];

/* İɽ */
typedef struct {
  int max_gone;				/* 14-Ǿİ */
  int quadrupling;			/* ʲǽ */
  int one_suit;				/* 쿧μ(ʣ) */
  int values[MAX_EVAL];			/* ɾ */
  hook_table_t hook;			/* ͭ */
} going_table_t;

#define refer_going(n)		(14 - going_table[n].max_gone)

extern going_table_t going_table[MAX_SORT];

typedef void (*receiver_t)(set_t *sets, int count);
typedef int (*chooser_t)(int pid, int n);

extern int evaluate(int pid, int n);

extern void generate_combinations(int *hand, receiver_t receiver);
extern int tell_completion_bonus(evaluated_hand_t *e);
extern int judge(int *hand, int *values, evaluated_hand_t *e);
extern int default_choose(int pid, int n);

extern int estimate_no_terminals(evaluated_hand_t *e);
extern int estimate_terminals(evaluated_hand_t *e);
extern int estimate_sequences(evaluated_hand_t *e);
extern int estimate_sisters(evaluated_hand_t *e);
extern int estimate_straight(evaluated_hand_t *e);
extern int estimate_3_suit_equals(evaluated_hand_t *e);
extern int estimate_1_suit(evaluated_hand_t *e);
extern int estimate_triplets(evaluated_hand_t *e);
extern int estimate_7_pairs(evaluated_hand_t *e);
extern int estimate_13_orphans(evaluated_hand_t *e);

extern int calculate_hand(int *hand, hook_table_t h);
extern int calculate_going(set_t *set, int count);
extern int calculate_going_7_pairs(int *hand, hook_table_t h);
extern int calculate_going_13_orphans(int *hand, hook_table_t h);
extern void detect_hooks(set_t *set, int count, int going, hook_table_t h);
extern int find_hook_types(set_t *set, int count, int last_tile);

extern int most_suit(int pid);

/*
 * 
 */

#define MAX_PLAYER		4

#define WIND_EAST		30	/*  */
#define WIND_SOUTH		31	/*  */
#define WIND_WEST		32	/*  */
#define WIND_NORTH		33	/*  */

#define current_discarder()	(__discarder)
#define current_dealer()	((roundkeeper.first_dealer + roundkeeper.rounds) % 4)
#define round_wind()		(roundkeeper.round_wind + WIND_EAST)
#define seat_wind_of(seat)	((4 + (seat) - current_dealer()) % 4 + WIND_EAST)

#define round_ended()		(roundkeeper.round_wind >= 2)

#define next_seat(s)		(((s) + 1) % 4)
#define prev_seat(s)		(((s) + 3) % 4)

typedef struct {
  int first_dealer;			/*  */
  int round_wind;			/*  */
  int rounds;				/*  */
  int sticks;				/* Ѥ */
  int carried_over[MAX_PLAYER];		/* ۤ꡼ */
} roundkeeper_t;

extern int __discarder;			/* β */
extern roundkeeper_t roundkeeper;
extern char round_caption[32];

extern void start_round(int first_dealer);

/*
 * 
 */

#define READY_EXPENSE		1000
#define STICK_GRATUITY		100
#define INITIAL_FUNDAGE		27000

#define MAX_STICK		4

typedef struct _hand_name_node {
  const char *name;			/* ̾ */
  int bonus;				/* ̿ */
  struct _hand_name_node *next;		/*  */
} hand_name_t;

typedef struct {
  hand_name_t *names;			/* ΩΥꥹ */
  const char *fmt;			/* ʸ "%d%d" */
  int point;				/*  */
  int bonus;				/* ̿ */
  int total;				/*  */
} scoresheet_t;

extern scoresheet_t *compute_score(int pid, int hands, int completions, int honors,
                                   int prizes, int point, int revealed);
extern void collect_money_each(int pid, int *bill);
extern void collect_money_loser(int pid, int loser, int *bill);
extern void fined_penalty(int pid, int *bill);

/*
 * (쥤)
 */

#define SURFACE_HEAD		0	/*  */
#define SURFACE_FACE		1	/* ɽ */
#define SURFACE_BACK		2	/*  */

#define MAX_SURFACE		3

#define MAX_DIRECTION		4

#define make_tid(n, d, s)	((n) | (d) << 8 | (s) << 16)
#define surface_of(tid)		((tid) >> 16)
#define direction_of(tid)	((tid) >> 8 & 0x0f)

#define toggle_angled(tid)	((tid) | 0x8000)
#define is_angled(tid)		((tid) & 0x8000)

#define toggle_disappear(tid)	((tid) | 0x4000)
#define is_disappear(tid)	((tid) & 0x4000)

typedef struct _bone_node {
  int tid;				/*  */
  int x;				/* ʿɸ */
  int y;				/* ľɸ */
  struct _bone_node *next;
} bone_t;

typedef struct {
  int		count;			/* פο */
  bone_t *	bones;			/* פ¤ */
  bone_t	__buf[16];		/* פ¤(> 14) */
} handrail_t;

extern bone_t *chain_bone(int tid, int x, int y, bone_t *next);
extern rectangle_t *bucket_append(int pid, int n, int angled);
extern rectangle_t *exposure_append(int pid, set_t set);
extern rectangle_t *exposure_append_quadruplet(int pid, int n);
extern rectangle_t *exposure_append_quadruplet_lifting(int pid, int n);
extern void align_handrail(int pid, int n, int surface);
extern void toggle_handrail_surface(int pid, bone_t *b, int surface);
extern bone_t *make_wall_bones(int *tomb);

/*
 * 
 */

#define COND_FIRST_TURN		0	/*  */
#define COND_READY		1	/* Ωľ */
#define COND_READY_DOUBLED	2	/* ֥Ωľ(Ωľ򽤾) */
#define COND_IMMEDIATE		3	/* ȯ */
#define COND_REVEALED		4	/* Ϫ */
#define COND_ONLY_DECLARED	5	/* Ωľ */
#define COND_NOT_YET_INDICATED	6	/* Ωľؼפ줿 */

#define is_ready(pid)		(bittest(players[pid].condition, COND_READY))

typedef struct {
  int		hand[MAX_SORT];		/*  */
  int		revealed_count;		/* Ϫ */
  set_t		revealed[4];		/* Ϫ */
  int		condition;		/*  */
  int		going;			/* İ */
  hook_table_t	winnings;		/* λ */
  hook_table_t	robbings;		/* Ĥ */
  rectangle_t *	discard[MAX_SORT];	/* Τ */
  rectangle_t *	through[MAX_SORT];	/* ̤ */
  int		dead_tiles[MAX_SORT];	/*  */
  scoresheet_t*	scoresheet;		/*  */

  handrail_t	handrail;		/* פ¤ */
  bone_t *	bucket;			/* Τפ¤ */
  bone_t *	exposure;		/* Ϫפ¤ */
} player_t;

extern player_t players[MAX_PLAYER];
extern int player_score[MAX_PLAYER];

extern int player_drawn(int pid, int n, int forbids_quadruplet);
extern int player_notified(int pid, int n, int permitted_types);

/*
 * 
 */

extern void pool_shuffle(void);
extern void pool_deal(void);
extern int pool_finished(void);
extern int pool_remain(void);
extern int pool_count_quadruplets(void);
extern int pool_draw(int pid);
extern int pool_draw_after_quadruplet(int pid);
extern int pool_discard(int pid, int n, int angled);
extern int pool_expose(int pid, int n);
extern void pool_expose_robbing(int pid, int n, int type, int whose);
extern int pool_expose_quadruplet(int pid, int n);
extern int count_prizes(int *hand, int ready);

extern bone_t *wall_bones;

/*
 * ʹ
 */

#include "event.h"

extern void dispatch(int event);

extern int batch_choose(int pid, int n, int forbids_quadruplet);
extern int batch_notified(int pid, int n, int permitted_types);

extern int interactive_choose(int pid, int n, int forbids_quadruplet);
extern int interactive_notified(int pid, int n, int permitted_types);
extern int interactive_report_winning(const char *reason, int pid, int n, int *bill);
extern int interactive_report_round_over(const char *reason, int *bill);

#endif /* mahjong_H */
