/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: player.c,v 1.2 1999/11/06 01:07:50 shinmaru Exp $
 */

#include "mahjong.h"
#include "option.h"

int batch_notified(int pid, int n, int permission)
{
  player_t *p = &players[pid];
  
  if (p->winnings[n].types != 0) {
    p->hand[n]++;
    evaluate(pid, n);
    p->hand[n]--;
    if (p->scoresheet != NULL)
      return reply_winning(pid, n);
  }
  return REPLY_NEXT;
}

int batch_choose(int pid, int n, int forbids_quadruplet)
{
  player_t *p = &players[pid];
  
  if (is_ready(pid)) {
    /* Ωľϼڤ */
    if (p->winnings[n].types != 0) {
      evaluate(pid, n);
      if (p->scoresheet != NULL)
        return event_winning(pid, n);
    }
  } else {
    evaluate(pid, n);
    if (p->scoresheet != NULL) {
      return event_winning(pid, n);
    } else {
      n = default_choose(pid, n);

      p->going = refer_going(n);

      if (p->going == 1) {
        /* λ */
        memcpy(p->winnings, going_table[n].hook, sizeof(hook_table_t));
        if (!bittest(p->condition, COND_REVEALED))
          return event_ready(pid, n);
      }
    }
  }
  if (!forbids_quadruplet && going_table[n].quadrupling)
    return event_quadruple(pid, n);
  else
    return event_discard(pid, n);
}

#define ALL (~(bit(HOOK_BELOW) | bit(HOOK_ABOVE)))

/* Ϫޥ */
static const int no_terminal_hooks[MAX_SORT] = {
  0, bit(HOOK_TRIPLET), ALL, ALL, ALL, ALL, ALL, bit(HOOK_TRIPLET), 0,
  0, bit(HOOK_TRIPLET), ALL, ALL, ALL, ALL, ALL, bit(HOOK_TRIPLET), 0,
  0, bit(HOOK_TRIPLET), ALL, ALL, ALL, ALL, ALL, bit(HOOK_TRIPLET), 0,
  0, 0, 0, 0, 0, 0, 0
};

/* ɾκ */
const int max_eval_values[MAX_EVAL] = {
  14,	/* 䥪 */
  14,	/*  */
  14,	/* ʿ */
   6,	/* ָ */
   9,	/* 쵤̴ */
   9,	/* Ʊ */
  14,	/* 쿧 */
   9,	/* С */
  14,	/* л */
  14	/* ̵ */
};

static int expect_hand(going_table_t *g)
{
  int max_value = INT_MIN, result = 0;
  int i;

  for (i = 0; i < MAX_EVAL; i++) {
    int value = g->values[i] - max_eval_values[i];

    if (max_value == value) {
      result |= bit(i);
    }
    else if (max_value < value) {
      max_value = value;
      result = bit(i);
    }
  }
  return result;
}

/* ׷ */
int default_choose(int pid, int drawn)
{
  player_t *p = &players[pid];
  int max_gone = 0, max_weight = 0;
  int result;
  int expection;
  int i;

  for (i = 0; i < MAX_SORT; i++) {
    if (max_gone < going_table[i].max_gone)
      max_gone = going_table[i].max_gone;
  }

  for (i = 0; i < MAX_SORT; i++) {
    if (going_table[i].max_gone == max_gone) {
      expection |= expect_hand(&going_table[i]);
    }
  }

  /* ǽιʤߡİȤʤפȴ */
  for (i = MAX_SORT - 1; i >= 0; i--) {
    int weight = 0;
    if (going_table[i].max_gone == max_gone) {
      int n;
      /* ͭʼˤĤĤ롣ξ8ĥ4ĥ6 */
      for (n = 0; n < MAX_SORT; n++) {
        int rest_count = 4 - p->dead_tiles[n];
        if (rest_count >= going_table[i].hook[n].needs) {
          weight += rest_count;
	  if (bittest(going_table[i].hook[n].types, HOOK_ABOVE))
	    weight += 4;
	  if (bittest(going_table[i].hook[n].types, HOOK_BELOW))
	    weight += 4;
	  if (bittest(going_table[i].hook[n].types, HOOK_MIDDLE))
	    weight += 6;
	}
      }
      /* פ˴Ԥû */
      for (n = 0; n < MAX_EVAL; n++) {
	if (bittest(expection, n))
	  weight += going_table[i].values[n];
      }
      if (max_weight < weight) {
        max_weight = weight;
        result = i;
      }
    }
  }

  if (option_stupid) {
    /* Ĥ */
    memcpy(p->robbings, going_table[result].hook, sizeof(hook_table_t));
  } else {
    going_table_t *g = &going_table[result];

    expection = expect_hand(g);

    /* ˤƤѤβǽ뤿ᡢϪե饰̵ˤ */
    memset(p->robbings, 0, sizeof(hook_table_t));

    if (bittest(expection, EVAL_1_SUIT) && g->max_gone - g->values[EVAL_1_SUIT] < 2) {
      /* Х򼭤ʤ: ޤǽʤ */
      int n;
      switch (g->one_suit) {
      case bit(SUIT_CHARACTER): n = 0; goto L1;
      case bit(SUIT_BALL): n = 9; goto L1;
      case bit(SUIT_BAMBOO): n = 18; L1:
	
	for (i = 0; i < 9; i++, n++) {
	  p->robbings[n].types = g->hook[n].types;
	}
	for (n = 27; n < MAX_SUIT; n++) {
	  p->robbings[n].types = g->hook[n].types;
	}
      }
    }

    if (bittest(expection, EVAL_TRIPLETS)) {
      for (i = 0; i < MAX_SUIT; i++)
	p->robbings[i].types = g->hook[i].types & HOOK_TRIPLET;
    }
  }
  
  return result;
}
