/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: pure.c,v 1.2 1999/11/06 01:07:50 shinmaru Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define random_integer(n)	(random() % (n))

/* ʤΤʤ */
static const char *tile_face[34] = {
  "", "", "", "", "", "ϻ", "", "Ȭ", "",
  "", "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", ""
};


int count(int *hand)
{
  int i, sum;
  for (i = 0, sum = 0; i < 10; i++)
    sum += hand[i];
  return sum;
}

int check_pair(int *used)
{
  int i;

  for (i = 0; i < 10; i++)
    if (used[i] != 0 && used[i] != 2)
      return 0;
  return 1;
}

int check_loop(int *used)
{
  int temp[10];
  int i;

  if (count(used) == 2)
    return check_pair(used);

  for (i = 0; i < 10; i++)
    temp[i] = used[i];

  for (i = 0; i < 10; i++) {
    if (i != 0 && i < 8 && temp[i] && temp[i + 1] && temp[i + 2]) {
      temp[i]--; temp[i + 1]--; temp[i + 2]--;
      if (check_loop(temp))
	return 1;
      temp[i]++; temp[i + 1]++; temp[i + 2]++;
    }
    if (temp[i] >= 3) {
      temp[i] -= 3;
      if (check_loop(temp))
	return 1;
      temp[i] += 3;
    }
  }
  return 0;
}

int check(int *hand)
{
  int used[10];
  int i;
  
  for (i = 0; i < 10; i++) used[i] = 0;

  for (i = 0; i < 14; i++)
    if (++used[hand[i]] > 4)
      return 0;

  if (check_pair(used))
    return 1;
  else
    return check_loop(used);
}

void print_tile(FILE *fp, int tile)
{
  if (tile == 0)
    fprintf(fp, "H");
  else
    fprintf(fp, "%d", tile);
}

void print_hand(void)
{
  int hand[14];
  int used[10], flags[10];
  int i, n;

  for (i = 0; i < 10; i++) used[i] = 0;
  
  n = random_integer(9) + 1;
  hand[0] = hand[1] = n;
  used[n] = 2;

  for (i = 2; i < 14; i +=3) {
  loop:
    /* Ϣ³0ˤʤ뤳Ȥʤ */
    if (random_integer(4) != 3) {
      n = random_integer(7) + 1;
      if (used[n] < 4 && used[n + 1] < 4 && used[n + 2] < 4) {
	hand[i] = n; hand[i + 1] = n + 1; hand[i + 2] = n + 2;
	used[n]++; used[n + 1]++; used[n + 2]++;
      } else
	goto loop;
    } else {
      n = random_integer(10);
      if (used[n] <= 1) {
	hand[i] = hand[i + 1] = hand[i + 2] = n;
	used[n] += 3;
      } else
	goto loop;
    }
  }

  for (i = 0; i < 10; i++) {
    hand[13] = i;
    flags[i] = check(hand);
  }

  for (i = 0; i < 13; i++) {
    int n = hand[i];
    if (n == 0) {
      fprintf(stdout, "H");
    } else {
      fprintf(stdout, "%d", n);
    }
  }
  fprintf(stdout, "\n");
  
  for (i = 1; i < 10; i++) {
    if (flags[i]) {
      fprintf(stderr, "%s", tile_face[i - 1]);
    }
  }
  if (flags[0]) {
    fprintf(stderr, "%s", tile_face[27]);
  }
  fprintf(stderr, "\n");
}

int main(int argc, char *argv[])
{
  srandom((unsigned)time(NULL));
  if (argc != 2) {
  usage:
    fprintf(stderr, "usage: pure <repeat-count>\n");
    return 1;
  } else {
    int i, n;

    n = atoi(argv[1]);
    if (n == 0)
      goto usage;

    for (i = 0; i < n; i++) {
      print_hand();
    }
    return 0;
  }
}
