/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: set.c,v 1.2 1999/11/06 01:07:51 shinmaru Exp $
 */

#include "mahjong.h"

#define is_terminal_tile(n)	(is_honor(n) || number_of(n) == 0 || number_of(n) == 8)

int tile_count_of_set[MAX_TYPEOF_SET] = { 1, 2, 2, 2, 3, 3, 4 };

int is_terminal(set_t set)
{
  int n = tile_of(set);
  switch (type_of(set)) {
  case CONJUNCT: return number_of(n) == 0 || number_of(n) == 7;
  case SPLITTER:
  case SEQUENCE: return number_of(n) == 0 || number_of(n) == 6;
  default: return is_terminal_tile(n);
  }
}

int is_green(set_t set)
{
  int n = tile_of(set);
  switch (type_of(set)) {
  case CONJUNCT: return n == 19 || n == 20;
  case SPLITTER:
  case SEQUENCE: return n == 19;
  default: return (19 <= n && n <= 21) || n == 23 || n == 25 || n == 28;
  }
}

int point_of(set_t set, int pid)
{
  int n = tile_of(set);
  switch (type_of(set)) {
  case PAIR:
    {
      int result = 0;
      if (is_dragon(n) || n == round_wind()) result += 2;
      if (n == seat_wind_of(pid)) result += 2;
      return result;
    }
  case TRIPLET:
    {
      int result = 4;
      if (is_revealed(set)) result /= 2;      
      if (is_terminal_tile(n)) result *= 2;
      return result;
    }
  case QUADRUPLET:
    {
      int result = 16;
      if (is_revealed(set)) result /= 2;
      if (is_terminal_tile(n)) result *= 2;
      return result;
    }
  }
  return 0;
}

void print_set(set_t set)
{
  int n = tile_of(set);
  putchar('[');
  switch (type_of(set)) {
  case SINGLETON:
    printf("%s", tile_face[n]);
    break;
  case PAIR:
    printf("%s%s", tile_face[n], tile_face[n]);
    break;
  case CONJUNCT:
    printf("%s%s", tile_face[n], tile_face[n + 1]);
    break;
  case SPLITTER:
    printf("%s%s", tile_face[n], tile_face[n + 2]);
    break;
  case SEQUENCE:
    switch (mark_of(set)) {
    case 1: printf("%s:%s%s", tile_face[n], tile_face[n + 1], tile_face[n + 2]); break;
    case 2: printf("%s:%s%s", tile_face[n + 1], tile_face[n], tile_face[n + 2]); break;
    case 3: printf("%s:%s%s", tile_face[n + 2], tile_face[n], tile_face[n + 1]); break;
    default: printf("%s%s%s", tile_face[n], tile_face[n + 1], tile_face[n + 2]);
    }
    break;
  case TRIPLET:
    switch (mark_of(set)) {
    case MARKING_RIGHT: printf(":"); break;
    case MARKING_FRONT: printf(":"); break;
    case MARKING_LEFT:  printf(":"); break;
    }
    printf("%s%s%s", tile_face[n], tile_face[n], tile_face[n]);
    break;
  case QUADRUPLET:
    if (is_revealed(set) && mark_of(set) == 0) {
      printf("%s%s", tile_face[n], tile_face[n]);
    } else {
      switch (mark_of(set)) {
      case MARKING_RIGHT: printf(":"); break;
      case MARKING_FRONT: printf(":"); break;
      case MARKING_LEFT:  printf(":"); break;
      }
      printf("%s%s%s%s", tile_face[n], tile_face[n], tile_face[n], tile_face[n]);
    }
  }
  putchar(']');
}
