/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: state.h,v 1.1.1.1 1999/11/05 01:55:18 shinmaru Exp $
 */

#ifndef state_H
#define state_H

typedef int event_t;

#define STATUS_FINAL		0	/* 켫 */
#define STATUS_AFTER_QUADRUPLET	1	/* 弫 */
#define STATUS_4_WINDS		2	/* Ϣ */
#define STATUS_4_QUADRUPLETS	3	/* ʻλ */
#define STATUS_4_PLAYERS_READY	4	/* ͲΩľ */

#define ACTION_SUSPEND		( 0 << 24)
#define ACTION_DEAL		( 1 << 24)
#define ACTION_DRAW		( 2 << 24)
#define ACTION_DISCARD		( 3 << 24)
#define ACTION_READY		( 4 << 24)
#define ACTION_WIN		( 5 << 24)
#define ACTION_ROB		( 6 << 24)
#define ACTION_QUADRUPLE	( 7 << 24)
#define ACTION_QUADRUPLE_LIFTING ( 8 << 24)
#define ACTION_AFTER_QUADRUPLET	( 9 << 24)
#define ACTION_ROUND_OVER	(10 << 24)
#define REPLY_WIN		(11 << 24)
#define REPLY_NEXT		(12 << 24)
#define REPORT_OK		(13 << 24)

#define action_of(e)		((e) & 0xff000000)
#define pid_of(e)		((e) >>  8 & 0xff)
#define permitted_types_of(e)	((e) >> 16 & 0xff)

#define event_drawing(p)	(ACTION_DRAW | (p) << 8)
#define event_discard(p, n)	(ACTION_DISCARD | (n) | (p) << 8)
#define event_ready(p, n)	(ACTION_READY | (n) | (p) << 8)
#define event_winning(p, n)	(ACTION_WIN | (n) | (p) << 8)
#define event_robbing(p, n, t)	(ACTION_ROB | (n) | (p) << 8 | (t) << 16)
#define event_quadruple(p, n)	(ACTION_QUADRUPLE | (n) | (p) << 8)
#define event_quadruple_lifting(p, n) (ACTION_QUADRUPLE_ADDING | (n) | (p) << 8)
#define event_after_quadruplet(p) (ACTION_AFTER_QUADRUPLET | (p) << 8)

#define reply_winning(p, n)	(REPLY_WIN | (n) | (p) << 8)
#define reply_next(p, n)	(REPLY_NEXT | (n) | (p) << 8)

extern void dispatch(int event);

extern void delay(int pid);
extern int interactive_choose(int pid, int n, int forbids_quadruplet);
extern int interactive_notified(int pid, int n, int permitted_types);
extern int interactive_report_winning(int pid, int n, scoresheet_t *scoresheet);
extern int interactive_report_round_over(const char *reason, int *details);

#endif /* state_H */
