/*-
 * Copyright (c) 1999 Morita Shinya.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: stub.c,v 1.2 1999/11/06 01:07:51 shinmaru Exp $
 */

#include "mahjong.h"
#include "debug.h"
#include "stub.h"

int interactive_choose(int pid, int n, int forbids_quadruplet)
{
  return batch_choose(pid, n, forbids_quadruplet);
}

int interactive_notified(int pid, int n, int permitted_types)
{
  return batch_notified(pid, n, permitted_types);
}

static void print_bill(int *bill)
{
  int i;
  for (i = 0; i < MAX_PLAYER; i++) {
    printf("\t%s %10d %+10d %10d\n", tile_face[seat_wind_of(i)],
	   player_score[i], bill[i], player_score[i] + bill[i]);
  }
}

int interactive_report_winning(const char *reason, int pid, int n, int *bill)
{
  printf("%s\n", round_caption);
  printf("%s\n", reason);

  if (pid != current_discarder()) {
    print_hand(pid);
    printf(" %s (%s)\n", tile_face[n], tile_face[seat_wind_of(current_discarder())]);
  }
  else {
    players[pid].hand[n]--;
    print_hand(pid);
    printf(" %s\n", tile_face[n]);
  }

  if (players[pid].scoresheet == NULL) {
    printf("@\n");
  } else {
    print_scoresheet(players[pid].scoresheet);
  }
  putchar('\n');
  if (bill != NULL) {
    print_bill(bill);
  }
  printf(".\n");
  return REPORT_OK;
}

int interactive_report_round_over(const char *reason, int *bill)
{
  printf("%s\n", round_caption);
  if (reason != NULL) {
    printf("%s\n", reason);
  }
  if (bill != NULL) {
    print_bill(bill);
  }
  printf(".\n");
  return REPORT_OK;
}

void repaint_bucket(int pid)
{
}

void repaint_handrail(int pid)
{
}

void repaint_wall(void)
{
}

void repaint_all(void)
{
}

void repaint_tomb(void)
{
}

void open_hand(int pid)
{
}

void open_wall(void)
{
}
