#include "windows/win32.h"
#include "mahjong.h"
#include "graphics.h"

typedef struct {
  char *bits;
  int width;
  int height;
  int scanline;
} image_t;

static image_t offscreen;

static image_t __hand[MAX_DIRECTION];
static image_t __face[MAX_DIRECTION];
static image_t __back[2];

static image_t __image[MAX_SURFACE][MAX_DIRECTION][MAX_SORT];
static image_t __stick[MAX_DIRECTION][MAX_STICK];
static image_t __yagamo[2];

static char *load_image(const char *filename, int width, int height)
{
  BITMAPFILEHEADER bmpfile;
  BITMAPINFOHEADER bmpinfo;
  FILE *fp;
  unsigned char *bits = NULL, *flip;
  int size;
  int i;

  fp = fopen(filename, "rb");
  if (fp == NULL)
    return NULL;

  if (fread(&bmpfile, sizeof(bmpfile), 1, fp) != 1 ||
      bmpfile.bfType != 0x4d42 ||
      fread(&bmpinfo, sizeof(bmpinfo), 1, fp) != 1)
    goto done;

  if (bmpinfo.biCompression != BI_RGB ||
      bmpinfo.biBitCount != 8 ||
      bmpinfo.biWidth != width || bmpinfo.biHeight != height)
    goto done;

  size = bmpinfo.biWidth * bmpinfo.biHeight;
  bits = (unsigned char *)malloc(size);
  flip = (unsigned char *)malloc(size);

  /* ᡼ɤ߹ */
  fseek(fp, bmpfile.bfOffBits, SEEK_SET);
  if (fread(flip, size, 1, fp) != 1) {
    free(bits);
    free(flip);
    bits = NULL;
    goto done;
  }

  /* 岼ȿž */
  for (i = 0; i < bmpinfo.biHeight; i++) {
    unsigned char *dst, *src;
    dst = bits + bmpinfo.biWidth * i;
    src = flip + bmpinfo.biWidth * (bmpinfo.biHeight - i - 1);
    memcpy(dst, src, bmpinfo.biWidth);
  }

  free(flip);
done:
  fclose(fp);
  return bits;
}

static int setup_images(void)
{
  struct {
    const char *filename;
    int width;
    int height;
    image_t *image;
  } img_src[] = {
    { "hand_0.bmp", 240, 168, &__hand[0] },
    { "hand_1.bmp",  20,  46, &__hand[1] },
    { "hand_2.bmp",  24,  42, &__hand[2] },
    { "hand_3.bmp",  20,  46, &__hand[3] },
    { "face_0.bmp", 240, 168, &__face[0] },
    { "face_1.bmp", 320, 140, &__face[1] },
    { "face_2.bmp", 240, 168, &__face[2] },
    { "face_3.bmp", 320, 140, &__face[3] },
    { "back_0.bmp",  24,  42, &__back[0] },
    { "back_1.bmp",  32,  35, &__back[1] },
    { "hstk_0.bmp",  72,   7, &__stick[0][0] },
    { "hstk_1.bmp",  72,   7, &__stick[0][1] },
    { "hstk_2.bmp",  72,   7, &__stick[0][2] },
    { "hstk_3.bmp",  72,   7, &__stick[0][3] },
    { "vstk_0.bmp",   8,  72, &__stick[1][0] },
    { "vstk_1.bmp",   8,  72, &__stick[1][1] },
    { "vstk_2.bmp",   8,  72, &__stick[1][2] },
    { "vstk_3.bmp",   8,  72, &__stick[1][3] },
    { "yagamo1.bmp", 40, 160, &__yagamo[0] },
    { "yagamo2.bmp", 40, 160, &__yagamo[1] }
  };
  int surface, d, n;
  char path[256];
  
  for (n = 0; n < sizeof(img_src)/sizeof(img_src[0]); n++) {
    image_t *img;
    
    sprintf(path, "..\\bmp\\%s", img_src[n].filename);

    img = img_src[n].image;
    img->width = img->scanline = img_src[n].width;
    img->height = img_src[n].height;
    img->bits = load_image(path, img->width, img->height);
    if (img->bits == NULL) {
      fprintf(stderr, "Can't load image: %s\n", path);
      return 1;
    }
  }

  /*  */
  for (n = 0; n < MAX_SORT; n++) {
    image_t *img = &__image[0][0][n];
    int x = n % 9 * image_size[0][0].width;
    int y = n / 9 * image_size[0][0].height;

    img->width = image_size[0][0].width;
    img->height = image_size[0][0].height;
    img->scanline = __hand[0].scanline;
    img->bits = __hand[0].bits + x + y * __hand[0].width;

    for (d = 1; d < MAX_DIRECTION; d++) {
      __image[0][d][n] = __hand[d];
    }
  }

  /* ɽ */
  for (d = 0; d < MAX_DIRECTION; d++) {
    for (n = 0; n < MAX_SORT; n++) {
      image_t *img = &__image[1][d][n];
      int x = n % 9 * image_size[1][d].width;
      int y = n / 9 * image_size[1][d].height;

      img->width = image_size[1][d].width;
      img->height = image_size[1][d].height;
      img->scanline = __face[d].scanline;
      img->bits = __face[d].bits + x + y * __face[d].width;
    }
  }
  
  /*  */
  for (n = 0; n < MAX_SORT; n++) {
    __image[2][0][n] = __back[0];
    __image[2][2][n] = __back[0];
    __image[2][1][n] = __back[1];
    __image[2][3][n] = __back[1];
  }

  /*  */
  for (n = 0; n < MAX_STICK; n++) {
    __stick[2][n] = __stick[0][n];
    __stick[3][n] = __stick[1][n];
  }
  
  return 0;
}

static int setup_offscreen(int width, int height)
{
  BITMAPINFO *bmi;
  BITMAPINFOHEADER *bi;
  int i;
  HDC dc;

  width = (width + 3) / 4 * 4;
  offscreen.width = width;
  offscreen.height = height;
  offscreen.scanline = width;

  bmi = (BITMAPINFO *)malloc(sizeof(BITMAPINFO) + 256 * sizeof(RGBQUAD));
  bi = &bmi->bmiHeader;
  bi->biSize          = sizeof(BITMAPINFOHEADER);
  bi->biWidth         = width;
  bi->biHeight        = -height;
  bi->biPlanes        = 1;
  bi->biBitCount      = 8;
  bi->biCompression   = BI_RGB;
  bi->biSizeImage     = 0;
  bi->biXPelsPerMeter = 0;
  bi->biYPelsPerMeter = 0;
  bi->biClrUsed       = 0;
  bi->biClrImportant  = 0;
  
  for (i = 0; i < 256; i++) {
    unsigned rgb = colormap[i];
    bmi->bmiColors[i].rgbRed      = (rgb & 0xff0000) >> 16;
    bmi->bmiColors[i].rgbGreen    = (rgb & 0x00ff00) >>  8;
    bmi->bmiColors[i].rgbBlue     = (rgb & 0x0000ff) >>  0;
    bmi->bmiColors[i].rgbReserved = 0;
  }

  dc = GetDC(0);
  offscreen_handle = CreateDIBSection(dc, bmi, DIB_RGB_COLORS,
                                      &(void *)offscreen.bits,
                                      NULL, 0);
  ReleaseDC(0, dc);
  free(bmi);
  
  return offscreen_handle == 0;
}

int setup_graphics(void)
{
  HDC dc;
  
  if (setup_images())
    return 1;

  if (setup_offscreen(layout_deck.width, layout_deck.height))
    return 1;

  dc = GetDC(main_window);
  
  offscreen_dc = CreateCompatibleDC(dc);
  SelectObject(offscreen_dc, offscreen_handle);

  ReleaseDC(main_window, dc);
  return 0;
}

void draw_image(image_t *d, int x1, int y1,
                image_t *s, int x2, int y2, int width, int height)
{
  unsigned char *dst, *src;
  int sscan, dscan;

  if (x1 < 0) { x2 -= x1; x1 = 0; }
  if (y1 < 0) { y2 -= y1; y1 = 0; }

  /* ºݤžȹ⤵ */
  width = min(width, min(d->width - x1, s->width - x2));
  height = min(height, min(d->height - y1, s->height - y2));

  if (width <= 0 || height <= 0)
    return;

  dst = d->bits + x1 + d->scanline * y1;
  src = s->bits + x2 + s->scanline * y2;
  dscan = d->scanline - width;
  sscan = s->scanline - width;

  while (height-- > 0) {
    int n = (width + 7) / 8;
    switch (width % 8) {
    case 0: do { *dst++ = *src++;
    case 7:	 *dst++ = *src++;
    case 6:	 *dst++ = *src++;
    case 5:	 *dst++ = *src++;
    case 4:	 *dst++ = *src++;
    case 3:	 *dst++ = *src++;
    case 2:	 *dst++ = *src++;
    case 1:	 *dst++ = *src++; } while (--n > 0);
    }
    dst += dscan;
    src += sscan;
  }
}

void draw_sprite(image_t *d, int x1, int y1,
                 image_t *s, int x2, int y2, int width, int height)
{
  unsigned char *dst, *src, mask;
  int sscan, dscan;

  if (x1 < 0) { x2 -= x1; x1 = 0; }
  if (y1 < 0) { y2 -= y1; y1 = 0; }

  /* ºݤžȹ⤵ */
  width = min(width, min(d->width - x1, s->width - x2));
  height = min(height, min(d->height - y1, s->height - y2));

  if (width <= 0 || height <= 0)
    return;

  dst = d->bits + x1 + d->scanline * y1;
  src = s->bits + x2 + s->scanline * y2;
  dscan = d->scanline - width;
  sscan = s->scanline - width;

  mask = *src;
  
  while (height-- > 0) {
    int n = (width + 7) / 8;
    switch (width % 8) {
    case 0: do { if (*src != mask) *dst = *src; dst++; src++;
    case 7:	 if (*src != mask) *dst = *src; dst++; src++;
    case 6:	 if (*src != mask) *dst = *src; dst++; src++;
    case 5:	 if (*src != mask) *dst = *src; dst++; src++;
    case 4:	 if (*src != mask) *dst = *src; dst++; src++;
    case 3:	 if (*src != mask) *dst = *src; dst++; src++;
    case 2:	 if (*src != mask) *dst = *src; dst++; src++;
    case 1:	 if (*src != mask) *dst = *src; dst++; src++; } while (--n > 0);
    }
    dst += dscan;
    src += sscan;
  }
}

static void draw_tile(int x, int y, int tid)
{
  image_t *img = &__image[surface_of(tid)][direction_of(tid)][tile_of(tid)];
  draw_sprite(&offscreen, x, y, img, 0, 0, img->width, img->height);
}

void put_tile(int x, int y, int tid)
{
  RECT rect;
  
  image_t *img = &__image[surface_of(tid)][direction_of(tid)][tile_of(tid)];
  draw_sprite(&offscreen, x, y, img, 0, 0, img->width, img->height);

  rect.left = x;
  rect.top = y;
  rect.right = x + img->width;
  rect.bottom = y + img->height;

  InvalidateRect(main_window, &rect, FALSE);
  SendMessage(main_window, WM_PAINT, 0, 0);
}

static void draw_bones(bone_t *b)
{
  for (; b != NULL; b = b->next) {
    if (!is_disappear(b->tid))
      draw_tile(b->x, b->y, b->tid);
  }
}

static void draw_bones_reverse(bone_t *b)
{
  if (b != NULL) {
    draw_bones_reverse(b->next);
    if (!is_disappear(b->tid))
      draw_tile(b->x, b->y, b->tid);
  }
}

void repaint_handrail(int pid)
{
  RECT rect;

  rectangle_to_RECT(&layout_hand[pid], &rect);
  
  FillRect(offscreen_dc, &rect, background_brush);

  switch (pid) {
  case 0:
    draw_bones(players[0].exposure);
    draw_bones(players[0].handrail.bones);
    break;
  case 1:
    draw_bones_reverse(players[1].exposure);
    draw_bones(players[1].handrail.bones);
    break;
  case 2:
    draw_bones_reverse(players[2].exposure);
    draw_bones(players[2].handrail.bones);
    break;
  case 3:
    draw_bones(players[3].exposure);
    draw_bones_reverse(players[3].handrail.bones);
    break;
  }

  InvalidateRect(main_window, &rect, FALSE);
  
  SendMessage(main_window, WM_PAINT, 0, 0);
}

void repaint_bucket(int pid)
{
  RECT rect;

  rectangle_to_RECT(&layout_bucket[pid], &rect);
  
  FillRect(offscreen_dc, &rect, background_brush);

  switch (pid) {
  case 0: draw_bones_reverse(players[0].bucket); break;
  case 1: draw_bones(players[1].bucket); break;
  case 2: draw_bones(players[2].bucket); break;
  case 3: draw_bones_reverse(players[3].bucket); break;
  }

  InvalidateRect(main_window, &rect, FALSE);

  if (players[pid].condition & (bit(COND_READY) | bit(COND_ONLY_DECLARED))) {
    image_t *img = &__stick[pid][1];
    
    rectangle_to_RECT(&layout_ready[pid], &rect);
    draw_image(&offscreen, rect.left, rect.top, img, 0, 0, img->width, img->height);
    
    InvalidateRect(main_window, &rect, FALSE);
  }
  
  SendMessage(main_window, WM_PAINT, 0, 0);
}

void repaint_wall(void)
{
  RECT rect;

  rectangle_to_RECT(&layout_wall, &rect);

  FillRect(offscreen_dc, &rect, background_brush);
  draw_bones_reverse(wall_bones);

  InvalidateRect(main_window, &rect, FALSE);

  SendMessage(main_window, WM_PAINT, 0, 0);
}


static void repaint_yagamo(void)
{
  int pid = roundkeeper.first_dealer;
  const rectangle_t *r = &layout_yagamo[pid];
  image_t *img = &__yagamo[min(1, roundkeeper.round_wind)];
  RECT rect;

  draw_image(&offscreen, r->x, r->y, img, 0, pid * r->height, r->width, r->height);

  rectangle_to_RECT(r, &rect);
  
  InvalidateRect(main_window, &rect, FALSE);

  SendMessage(main_window, WM_PAINT, 0, 0);
}

void repaint_all(void)
{
  RECT rect;
  int i;

  rectangle_to_RECT(&layout_deck, &rect);
  FillRect(offscreen_dc, &rect, background_brush);

  repaint_yagamo();

  repaint_wall();

  for (i = 0; i < MAX_PLAYER; i++) {
    repaint_bucket(i);
    repaint_handrail(i);
  }

  InvalidateRect(main_window, &rect, FALSE);

  SendMessage(main_window, WM_PAINT, 0, 0);
}

void set_highlight(rectangle_t *r)
{
  RECT rect;

  if (highlight != NULL) {
    rectangle_to_RECT(highlight, &rect);
    InvalidateRect(main_window, &rect, FALSE);    
  }
  highlight = r;
  if (highlight != NULL) {
    rectangle_to_RECT(highlight, &rect);
    InvalidateRect(main_window, &rect, FALSE);    
  }
  
  SendMessage(main_window, WM_PAINT, 0, 0);
}

void set_screening(rectangle_t *r)
{
  RECT rect;

  if (screening != NULL) {
    rectangle_to_RECT(screening, &rect);
    InvalidateRect(main_window, &rect, FALSE);    
  }
  screening = r;
  if (screening != NULL) {
    rectangle_to_RECT(screening, &rect);
    InvalidateRect(main_window, &rect, FALSE);    
  }
  
  SendMessage(main_window, WM_PAINT, 0, 0);
}

void set_selection(rectangle_t *r)
{
  RECT rect;
  
  if (selection != NULL) {
    rectangle_to_RECT(selection, &rect);
    InvalidateRect(main_window, &rect, FALSE);    
  }
  selection = r;
  if (selection != NULL) {
    rectangle_to_RECT(selection, &rect);
    InvalidateRect(main_window, &rect, FALSE);    
  }

  SendMessage(main_window, WM_PAINT, 0, 0);
}

void show_selection(void)
{
  POINT p;
  
  GetCursorPos(&p);
  ScreenToClient(main_window, &p);
  
  pointer_move(p.x, p.y);
}

void clear_rectangle(int x, int y, int width, int height)
{
  RECT rect;

  rect.left = x;
  rect.top = y;
  rect.right = x + width;
  rect.bottom = y + height;
  
  FillRect(offscreen_dc, &rect, background_brush);

  InvalidateRect(main_window, &rect, FALSE);

  SendMessage(main_window, WM_PAINT, 0, 0);
}

void beep(void)
{
  MessageBeep(0);
}

void open_hand(int pid)
{
  bone_t *b;
  for (b = players[pid].handrail.bones; b != NULL; b = b->next) {
    if (surface_of(b->tid) != SURFACE_FACE)
      toggle_handrail_surface(pid, b, SURFACE_FACE);
  }
  repaint_handrail(pid);
}

void open_wall(void)
{
  bone_t *b;

  for (b = wall_bones; b != NULL; b = b->next) {
    if (surface_of(b->tid) == SURFACE_FACE && b->next != NULL) {
      b->y += 48;
      b = b->next;
      b->tid = make_tid(tile_of(b->tid), 0, SURFACE_FACE);
    }
  }
  repaint_wall();
}
