#include "windows/win32.h"
#include "mahjong.h"
#include "option.h"
#include "graphics.h"

#include "windows/yagamo.rc"

static HINSTANCE __instance;

HWND main_window;
HBITMAP offscreen_handle;
HDC offscreen_dc;

HBRUSH background_brush;
static HBRUSH highlight_brush;
static HBRUSH selection_brush;

rectangle_t *highlight;
rectangle_t *screening;
rectangle_t *selection;

static dimension_t window_size;
static HWND dialog;

static int register_window_classes(HINSTANCE prev);
static void version_dialog(void);
static LRESULT CALLBACK main_wnd_proc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam);

int WINAPI WinMain(HINSTANCE instance, HINSTANCE prev, LPSTR cmd, int cmd_show)
{
  MSG msg;

  option_silence = 1;
  option_wait = 1;
  random_seed = (unsigned)time(NULL);

  srand(random_seed);

  if (register_window_classes(instance)) {
    fprintf(stderr, "Can't register window classes.\n");
    return 1;
  }

  window_size.width = layout_deck.width + GetSystemMetrics(SM_CXFRAME) * 2;
  window_size.height = layout_deck.height + GetSystemMetrics(SM_CYFRAME) * 2
    + GetSystemMetrics(SM_CYCAPTION) - 1 + GetSystemMetrics(SM_CYMENU);

  main_window = CreateWindow("main_window", "Yagamo",
                             WS_OVERLAPPEDWINDOW & ~WS_MAXIMIZEBOX,
                             CW_USEDEFAULT, CW_USEDEFAULT,
                             window_size.width, window_size.height,
                             NULL,
                             NULL,
                             __instance, NULL);
  if (main_window == NULL) {
    fprintf(stderr, "Can't create main window.\n");
    return 1;
  }

  if (setup_graphics()) {
    fprintf(stderr, "setup_graphics() failed.\n");
    return 1;
  }

  background_brush = CreateSolidBrush(RGB(0, 128, 0));
  highlight_brush = CreateSolidBrush(RGB(255, 255, 0));
  selection_brush = CreateSolidBrush(RGB(0, 0, 128));

  ShowWindow(main_window, cmd_show);
  UpdateWindow(main_window);

  while (GetMessage(&msg, NULL, 0, 0)) {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }

  return msg.wParam;
}


int register_window_classes(HINSTANCE instance)
{
  WNDCLASSEX wc;

  /* ᥤ󥦥ɥ饹 */
  wc.cbSize        = sizeof(wc);
  wc.style         = 0;
  wc.lpfnWndProc   = (WNDPROC)main_wnd_proc;
  wc.cbClsExtra	   = 0;
  wc.cbWndExtra	   = 0;
  wc.hInstance	   = instance;
  wc.hIcon	   = NULL /*LoadIcon(instance, MAKEINTRESOURCE(IDI_ICON1))*/;
  wc.hCursor	   = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = NULL;
  wc.lpszMenuName  = MAKEINTRESOURCE(IDM_MENU1);
  wc.lpszClassName = "main_window";
  wc.hIconSm	   = NULL;

  return RegisterClassEx(&wc) == 0;
}

LRESULT CALLBACK main_wnd_proc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
  switch (msg) {
  case WM_CREATE:
    return 0;

  case WM_DESTROY:
    DeleteDC(offscreen_dc);
    DeleteObject(offscreen_handle);
    DeleteObject(background_brush);
    DeleteObject(highlight_brush);
    DeleteObject(selection_brush);
    PostQuitMessage(0);
    return 0;

  case WM_GETMINMAXINFO:
    {
      MINMAXINFO *mmi = (MINMAXINFO *)lparam;
      mmi->ptMaxTrackSize.x = window_size.width;
      mmi->ptMaxTrackSize.y = window_size.height;
      mmi->ptMinTrackSize.x = window_size.width;
      mmi->ptMinTrackSize.y = window_size.height;
    }
    return 0;

  case WM_NCHITTEST:
    {
      int ht = DefWindowProc(wnd, msg, wparam, lparam);
      switch (ht) {
      case HTTOPLEFT:	case HTTOP:	case HTTOPRIGHT:
      case HTLEFT:			case HTRIGHT:
      case HTBOTTOMLEFT:case HTBOTTOM:	case HTBOTTOMRIGHT:        
        ht = HTBORDER;
      }
      return ht;
    }
    
  case WM_PAINT:
  paint_window:
    {
      PAINTSTRUCT ps;
      HDC dc;
      HBRUSH hb_old;
      int x, y, width, height;

      dc = BeginPaint(wnd, &ps);

      x = ps.rcPaint.left;
      y = ps.rcPaint.top;
      width = ps.rcPaint.right  - x;
      height = ps.rcPaint.bottom - y;

      /* ե꡼󤫤饹꡼ž */
      BitBlt(dc, x, y, width, height, offscreen_dc, x, y, SRCCOPY);

      if (highlight != NULL) {
        hb_old = SelectObject(dc, highlight_brush);
        PatBlt(dc,
               highlight->x, highlight->y, highlight->width,
               highlight->height, PATINVERT);
        SelectObject(dc, hb_old);
      }
      if (selection != NULL) {
        hb_old = SelectObject(dc, selection_brush);
        PatBlt(dc,
               selection->x, selection->y, selection->width,
               selection->height, PATINVERT);
        SelectObject(dc, hb_old);
      }
      if (screening != NULL) {
        RECT rect;
        rectangle_to_RECT(screening, &rect);
        InvertRect(dc, &rect);
      }
      
      EndPaint(wnd, &ps);
    }
    return 0;    
  }
  
  if (dialog != NULL) {
    if (msg == WM_LBUTTONDOWN || msg == WM_CLOSE) {
      EnableWindow(wnd, FALSE);
      ShowWindow(dialog, SW_SHOW);
      return 0;
    }
    
    return DefWindowProc(wnd, msg, wparam, lparam);
  }
  
  switch (msg) {
  case WM_COMMAND:
    {
      int id = LOWORD(wparam);
      switch (id) {
      case CM_START:
        option_batch = 0;
        option_wait = 1;
        highlight = NULL;
        selection = NULL;
        screening = NULL;
        start_round(random_integer(4));
        break;

      case CM_BATCH:
        option_batch = 1;
        option_wait = 0;
        highlight = NULL;
        selection = NULL;
        screening = NULL;
        start_round(random_integer(4));
        break;
        
      case CM_QUIT:
        PostMessage(wnd, WM_DESTROY, 0, 0);
        break;
        
      case CM_HELP:
        break;
        
      case CM_ABOUT:
        version_dialog();
        break;
      }
      return 0;
    }

  case WM_LBUTTONDOWN:
    {
      short x = LOWORD(lparam);
      short y = HIWORD(lparam);
      if ((wparam & MK_CONTROL) == 0)
        button1_down(x, y);
      else
        button2_down(x, y);
    }
    return 0;

  case WM_MBUTTONDOWN:
    {
      short x = LOWORD(lparam);
      short y = HIWORD(lparam);
      button2_down(x, y);
    }
    return 0;

  case WM_RBUTTONDOWN:
    {
      short x = LOWORD(lparam);
      short y = HIWORD(lparam);
      button3_down(x, y);
    }
    return 0;

  case WM_MOUSEMOVE:
    {
      short x = LOWORD(lparam);
      short y = HIWORD(lparam);
      pointer_move(x, y);
    }
    return 0;
  }

  return DefWindowProc(wnd, msg, wparam, lparam);
}

static const char *__text;
static const int *__bill;

static void center_dialog(HWND wnd)
{
  int x, y, width, height;
  RECT rect;
  
  GetWindowRect(wnd, &rect);
  
  width = rect.right - rect.left;
  height = rect.bottom - rect.top;
  x = rect.left + (layout_deck.width - width) / 2;
  y = rect.top + (layout_deck.height - height) /2;
  
  MoveWindow(wnd, x, y, width, height, FALSE);
}

/* ץ1 */
BOOL CALLBACK dialog_proc1(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
  switch (msg) {
  case WM_INITDIALOG:
    dialog = wnd; /* ɬ */
    center_dialog(wnd);
    break;
    
  case WM_COMMAND:
    if (LOWORD(wparam) == IDC_BUTTON1) {
      dialog = NULL; /* ɬ */
      EndDialog(wnd, wparam);
    }
    else if (LOWORD(wparam) == IDC_BUTTON2) {
      EnableWindow(main_window, TRUE);
      ShowWindow(wnd, SW_HIDE);
    }
    return TRUE;
    
  case WM_DRAWITEM:
    {
      LPDRAWITEMSTRUCT d = (LPDRAWITEMSTRUCT)lparam;
      HDC dc = d->hDC;
      int mode_old;
      RECT rect;
      int y;
      const char *top, *end;

      FillRect(dc, &d->rcItem, GetStockObject(WHITE_BRUSH));

      mode_old = SetBkMode(dc, TRANSPARENT);
      TextOut(dc, 0, 0, round_caption, strlen(round_caption));

      top = end = __text;
      y = 20;
      for (;;) {
        int c = *end++;
        if (c == '\0' || c == '\n') {
          int bytes = end - top - 1;
          if (bytes > 0) {
            TextOut(dc, 8, y, top, bytes);
          }
          y += 20;
          top = end;
        }
        if (c == '\0')
          break;
      }

      SetBkMode(dc, mode_old);
    }
    return TRUE;
  }
  return FALSE;
}

/* ץ2 */
BOOL CALLBACK dialog_proc2(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
  switch (msg) {
  case WM_INITDIALOG:
    dialog = wnd; /* ɬ */
    center_dialog(wnd);
    break;
    
  case WM_COMMAND:
    if (LOWORD(wparam) == IDC_BUTTON1) {
      dialog = NULL; /* ɬ */
      EndDialog(wnd, wparam);
    }
    else if (LOWORD(wparam) == IDC_BUTTON2) {
      EnableWindow(main_window, TRUE);
      ShowWindow(wnd, SW_HIDE);
    }
    return TRUE;
    
  case WM_DRAWITEM:
    {
      LPDRAWITEMSTRUCT d = (LPDRAWITEMSTRUCT)lparam;
      HDC dc = d->hDC;
      int mode_old;
      char buf[256];
      int x, y;
      int i;
      /* ñ */
      static const rectangle_t dialog_bases[MAX_PLAYER] = {
        { 34, 76, 60, 32 },
        { 68, 42, 60, 32 },
        { 34,  8, 60, 32 },
        {  0, 42, 60, 32 }
      };
      RECT rect[MAX_PLAYER];

      for (i = 0; i < MAX_PLAYER; i++) {
        rectangle_to_RECT(&dialog_bases[i], &rect[i]);
        MapDialogRect(wnd, &rect[i]);
        FillRect(dc, &rect[i], GetStockObject(WHITE_BRUSH));
      }

      mode_old = SetBkMode(dc, TRANSPARENT);

      TextOut(dc, 0, 0, round_caption, strlen(round_caption));

      for (i = 0; i < MAX_PLAYER; i++) {
        int length;
        SIZE e;        
        
        x = rect[i].left;
        y = rect[i].top;
        
        length = sprintf(buf, "%s", tile_face[seat_wind_of(i)]);
        GetTextExtentPoint32(dc, buf, length, &e);
        TextOut(dc, x, y, buf, length);
        
        length = sprintf(buf, "%12d", player_score[i]);
        y += e.cy;
        TextOut(dc, x, y, buf, length);
        
        length = sprintf(buf, __bill[i] != 0 ? "%+12d" : "%12d", __bill[i]);
        y += e.cy;
        TextOut(dc, x, y, buf, length);
        
        length = sprintf(buf, "%12d", player_score[i] + __bill[i]);
        y += e.cy;
        TextOut(dc, x, y, buf, length);
      }
      
      SetBkMode(dc, mode_old);
    }
    return TRUE;
  }
  return FALSE;
}

void report_dialog(const char *text)
{
  if (text != NULL) {
    __text = text;
    DialogBox(__instance, MAKEINTRESOURCE(IDD_DIALOG1), main_window, dialog_proc1);
  }
}

void score_dialog(const int *bill)
{
  if (bill != NULL) {
    __bill = bill;
    DialogBox(__instance, MAKEINTRESOURCE(IDD_DIALOG1), main_window, dialog_proc2);
  }
}

/* ץ3 */
BOOL CALLBACK dialog_proc3(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
  switch (msg) {
  case WM_INITDIALOG:
    center_dialog(wnd);
    break;
    
  case WM_COMMAND:
    if (LOWORD(wparam) == IDOK) {
      EndDialog(wnd, wparam);
    }
    return TRUE;
  }
  return FALSE;
}

void version_dialog(void)
{
  DialogBox(__instance, MAKEINTRESOURCE(IDD_DIALOG2), main_window, dialog_proc3);
}